"""
Utility functions for photometry.
"""

import spectres
import numpy as np

from species.core import box
from species.read import read_model, read_calibration, read_filter, read_planck


def multi_photometry(datatype,
                     spectrum,
                     filters,
                     parameters):
    """
    Parameters
    ----------
    datatype : str
        Data type ('model' or 'calibration').
    spectrum : str
        Spectrum name (e.g., 'drift-phoenix').
    filters : tuple(str, )
        Filter IDs.
    parameters : dict
        Parameters and values for the spectrum

    Returns
    -------
    species.core.box.SynphotBox
        Box with synthetic photometry.
    """

    print('Calculating synthetic photometry...', end='', flush=True)

    flux = {}

    if datatype == 'model':
        for item in filters:
            if spectrum == 'planck':
                readmodel = read_planck.ReadPlanck(filter_name=item)
            else:
                readmodel = read_model.ReadModel(spectrum, filter_name=item)

            flux[item] = readmodel.get_flux(parameters)

    elif datatype == 'calibration':
        for item in filters:
            readcalib = read_calibration.ReadCalibration(spectrum, filter_name=item)
            flux[item] = readcalib.get_flux(parameters)

    print(' [DONE]')

    return box.create_box('synphot', name='synphot', flux=flux)


def apparent_to_absolute(app_mag,
                         distance):
    """
    Parameters
    ----------
    app_mag : float or numpy.ndarray
        Apparent magnitude (mag).
    distance : float or numpy.ndarray
        Distance (pc).

    Returns
    -------
    float or numpy.ndarray
        Absolute magnitude (mag).
    """

    return app_mag - 5.*np.log10(distance) + 5.


def get_residuals(datatype,
                  spectrum,
                  parameters,
                  filters,
                  objectbox,
                  inc_phot=True,
                  inc_spec=False):
    """
    Parameters
    ----------
    datatype : str
        Data type ('model' or 'calibration').
    spectrum : str
        Name of the atmospheric model or calibration spectrum.
    parameters : dict
        Parameters and values for the spectrum
    filters : tuple(str, )
        Filter IDs. All available photometry of the object is used if set to None.
    objectbox : species.core.box.ObjectBox
        Box with the photometry and/or spectrum of an object.
    inc_phot : bool
        Include photometry.
    inc_spec : bool
        Include spectrum.

    Returns
    -------
    species.core.box.ResidualsBox
        Box with the photometry and/or spectrum residuals.
    """

    if filters is None:
        filters = objectbox.filters

    if inc_phot:
        model_phot = multi_photometry(datatype=datatype,
                                      spectrum=spectrum,
                                      filters=filters,
                                      parameters=parameters)

        res_phot = np.zeros((2, len(objectbox.flux)))

        for i, item in enumerate(filters):
            transmission = read_filter.ReadFilter(item)

            res_phot[0, i] = transmission.mean_wavelength()
            res_phot[1, i] = (objectbox.flux[item][0]-model_phot.flux[item])/objectbox.flux[item][1]

    else:
        res_phot = None

    print('Calculating residuals...', end='', flush=True)

    if inc_spec:
        res_spec = {}

        for key in objectbox.spectrum:
            wavel_range = (0.9*objectbox.spectrum[key][0][0, 0],
                           1.1*objectbox.spectrum[key][0][-1, 0])

            if spectrum == 'planck':
                readmodel = read_planck.ReadPlanck(wavel_range=wavel_range)
                model = readmodel.get_spectrum(model_param=parameters, spec_res=1000.)

            else:
                readmodel = read_model.ReadModel(spectrum, wavel_range=wavel_range)
                model = readmodel.get_model(parameters, spec_res=None)

            wl_new = objectbox.spectrum[key][0][:, 0]

            flux_new = spectres.spectres(new_spec_wavs=wl_new,
                                         old_spec_wavs=model.wavelength,
                                         spec_fluxes=model.flux,
                                         spec_errs=None)

            res_tmp = (objectbox.spectrum[key][0][:, 1]-flux_new)/objectbox.spectrum[key][0][:, 2]

            res_spec[key] = np.column_stack([wl_new, res_tmp])

    else:
        res_spec = None

    print(' [DONE]')

    return box.create_box(boxtype='residuals',
                          name=objectbox.name,
                          photometry=res_phot,
                          spectrum=res_spec)



"""
Module for plotting MCMC results.
"""

import os

import corner
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from species.data import database
from species.util import plot_util


mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
mpl.rcParams['font.family'] = 'serif'

plt.rc('axes', edgecolor='black', linewidth=2.5)


def plot_walkers(tag,
                 nsteps=None,
                 offset=None,
                 output='walkers.pdf'):
    """
    Function to plot the step history of the walkers.

    Parameters
    ----------
    tag : str
        Database tag with the MCMC samples.
    nsteps : int
        Number of steps.
    offset : tuple(float, float)
        Offset of the x- and y-axis label.
    output : str
        Output filename.

    Returns
    -------
    None
    """

    print(f'Plotting walkers: {output}...', end='', flush=True)

    species_db = database.Database()
    box = species_db.get_samples(tag)

    samples = box.samples
    labels = plot_util.update_labels(box.parameters)

    ndim = samples.shape[-1]

    plt.figure(1, figsize=(6, ndim*1.5))
    gridsp = mpl.gridspec.GridSpec(ndim, 1)
    gridsp.update(wspace=0, hspace=0.1, left=0, right=1, bottom=0, top=1)

    for i in range(ndim):
        ax = plt.subplot(gridsp[i, 0])

        if i == ndim-1:
            ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                           direction='in', width=1, length=5, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=True)

            ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                           direction='in', width=1, length=3, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=True)

        else:
            ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                           direction='in', width=1, length=5, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=False)

            ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                           direction='in', width=1, length=3, labelsize=12, top=True,
                           bottom=True, left=True, right=True, labelbottom=False)

        if i == ndim-1:
            ax.set_xlabel('Step number', fontsize=10)
        else:
            ax.set_xlabel('', fontsize=10)

        ax.set_ylabel(labels[i], fontsize=10)

        if offset is not None:
            ax.get_xaxis().set_label_coords(0.5, offset[0])
            ax.get_yaxis().set_label_coords(offset[1], 0.5)

        else:
            ax.get_xaxis().set_label_coords(0.5, -0.22)
            ax.get_yaxis().set_label_coords(-0.09, 0.5)

        if nsteps is not None:
            ax.set_xlim(0, nsteps)

        for j in range(samples.shape[0]):
            ax.plot(samples[j, :, i], ls='-', lw=0.5, color='black', alpha=0.5)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')


def plot_posteriors(tag,
                    burnin=None,
                    title=None,
                    offset=None,
                    title_fmt='.2f',
                    limits=None,
                    output='posterior.pdf'):
    """
    Function to plot the posterior distributions.

    Parameters
    ----------
    tag : str
        Database tag with the MCMC samples.
    burnin : int, None
        Number of burnin steps to exclude. All samples are used if set to None.
    title : str, None
        Plot title.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    title_fmt : str
        Format of the median and error values.
    limits : tuple(tuple(float, float), ), None
        Axis limits of all parameters. Automatically set if set to None.
    output : str
        Output filename.

    Returns
    -------
    None
    """

    print(f'Plotting posteriors: {output}...', end='', flush=True)

    species_db = database.Database()
    box = species_db.get_samples(tag, burnin=burnin)

    samples = box.samples
    par_val = tuple(box.prob_sample.values())

    labels = plot_util.update_labels(box.parameters)

    ndim = samples.shape[-1]
    samples = samples.reshape((-1, ndim))

    fig = corner.corner(samples, labels=labels, quantiles=[0.16, 0.5, 0.84],
                        label_kwargs={'fontsize': 13}, show_titles=True,
                        title_kwargs={'fontsize': 12}, title_fmt=title_fmt)

    axes = np.array(fig.axes).reshape((ndim, ndim))

    for i in range(ndim):
        for j in range(ndim):
            if i >= j:
                ax = axes[i, j]

                ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                               direction='in', width=1, length=5, labelsize=12, top=True,
                               bottom=True, left=True, right=True)

                ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                               direction='in', width=1, length=3, labelsize=12, top=True,
                               bottom=True, left=True, right=True)

                if limits is not None:
                    ax.set_xlim(limits[j])

                ax.axvline(par_val[j], color='tomato')

                if i > j:
                    ax.axhline(par_val[i], color='tomato')
                    ax.plot(par_val[j], par_val[i], 's', color='tomato')

                    if limits is not None:
                        ax.set_ylim(limits[i])

                if offset is not None:
                    ax.get_xaxis().set_label_coords(0.5, offset[0])
                    ax.get_yaxis().set_label_coords(offset[1], 0.5)

                else:
                    ax.get_xaxis().set_label_coords(0.5, -0.26)
                    ax.get_yaxis().set_label_coords(-0.27, 0.5)

    if title:
        fig.suptitle(title, y=1.02, fontsize=16)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')


def plot_photometry(tag,
                    filter_id,
                    burnin=None,
                    xlim=None,
                    output='photometry.pdf'):
    """
    Function to plot the posterior distribution of the synthetic photometry.

    Parameters
    ----------
    tag : str
        Database tag with the MCMC samples.
    filter_id : str
        Filter ID.
    burnin : int, None
        Number of burnin steps to exclude. All samples are used if set to None.
    xlim : tuple(float, float), None
        Axis limits. Automatically set if set to None.
    output : strr
        Output filename.

    Returns
    -------
    None
    """

    species_db = database.Database()

    samples = species_db.get_mcmc_photometry(tag, burnin, filter_id)

    print(f'Plotting photometry samples: {output}...', end='', flush=True)

    fig = corner.corner(samples, labels=['Magnitude'], quantiles=[0.16, 0.5, 0.84],
                        label_kwargs={'fontsize': 13}, show_titles=True,
                        title_kwargs={'fontsize': 12}, title_fmt='.2f')

    axes = np.array(fig.axes).reshape((1, 1))

    ax = axes[0, 0]

    ax.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                   direction='in', width=1, length=5, labelsize=12, top=True,
                   bottom=True, left=True, right=True)

    ax.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                   direction='in', width=1, length=3, labelsize=12, top=True,
                   bottom=True, left=True, right=True)

    if xlim is not None:
        ax.set_xlim(xlim)

    ax.get_xaxis().set_label_coords(0.5, -0.26)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print(' [DONE]')



"""
Text
"""

import os
import urllib.request

import h5py
import numpy as np

from astropy.io import fits

from species.data import queries
from species.util import data_util


def add_vlm_plx(input_path,
                database):
    """
    Function for adding the Database of Ultracool Parallaxes to the database.

    Parameters
    ----------
    input_path : str
    database : h5py._hl.files.File

    Returns
    -------
    NoneType
        None
    """

    data_file = os.path.join(input_path, 'vlm-plx-all.fits')

    url = 'http://www.as.utexas.edu/~tdupuy/plx/' \
          'Database_of_Ultracool_Parallaxes_files/vlm-plx-all.fits'

    if not os.path.isfile(data_file):
        print('Downloading Database of Ultracool Parallaxes (307 kB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Adding Database of Ultracool Parallaxes...', end='', flush=True)

    group = 'photometry/vlm-plx'

    database.create_group(group)

    hdulist = fits.open(data_file)
    photdata = hdulist[1].data

    plx = photdata['PLX']  # [mas]
    distance = 1./(plx*1e-3)  # [pc]

    name = photdata['NAME']
    name = np.core.defchararray.strip(name)

    sptype = photdata['ISPTSTR']
    sptype = np.core.defchararray.strip(sptype)

    sptype_op = photdata['OSPTSTR']
    sptype_op = np.core.defchararray.strip(sptype_op)

    for i, item in enumerate(sptype):
        if item == 'null':
            sptype[i] = sptype_op[i]

    flag = photdata['FLAG']
    flag = np.core.defchararray.strip(flag)

    sptype = data_util.update_sptype(sptype)

    dtype = h5py.special_dtype(vlen=bytes)

    dset = database.create_dataset(group+'/name', (np.size(name), ), dtype=dtype)
    dset[...] = name

    dset = database.create_dataset(group+'/sptype', (np.size(sptype), ), dtype=dtype)
    dset[...] = sptype

    dset = database.create_dataset(group+'/flag', (np.size(flag), ), dtype=dtype)
    dset[...] = flag

    database.create_dataset(group+'/distance', data=distance)
    database.create_dataset(group+'/MKO/NSFCam.Y', data=photdata['YMAG'])
    database.create_dataset(group+'/MKO/NSFCam.J', data=photdata['JMAG'])
    database.create_dataset(group+'/MKO/NSFCam.H', data=photdata['HMAG'])
    database.create_dataset(group+'/MKO/NSFCam.K', data=photdata['KMAG'])
    database.create_dataset(group+'/MKO/NSFCam.Lp', data=photdata['LMAG'])
    database.create_dataset(group+'/MKO/NSFCam.Mp', data=photdata['MMAG'])
    database.create_dataset(group+'/2MASS/2MASS.J', data=photdata['J2MAG'])
    database.create_dataset(group+'/2MASS/2MASS.H', data=photdata['H2MAG'])
    database.create_dataset(group+'/2MASS/2MASS.Ks', data=photdata['K2MAG'])

    print(' [DONE]')

    # print('Querying SIMBAD...', end=')

    # simbad_id = queries.get_simbad(name)

    # dset = database.create_dataset(group+'/simbad', (np.size(simbad_id), ), dtype=dtype)
    # dset[...] = simbad_id

    # print(' [DONE]')

    database.close()



"""
Utility functions for running the unit tests.
"""

import os


def create_config(test_path):
    """
    Function for creating a configuration file in the test folder.

    Parameters
    ----------
    test_path : str
        Folder where the unit tests are located.

    Returns
    -------
    NoneType
        None
    """

    config_file = os.path.join(test_path, 'species_config.ini')
    database_file = os.path.join(test_path, 'species_database.hdf5')
    data_folder = os.path.join(test_path, 'data/')

    with open(config_file, 'w') as config:
        config.write('[species]\n')
        config.write('database = '+database_file+'\n')
        config.write('data_folder = '+data_folder)



"""
Module for DRIFT-PHOENIX atmospheric model spectra.
"""

import os
import tarfile
import urllib.request

import numpy as np

from species.util import data_util


def add_drift_phoenix(input_path,
                      database):
    """
    Function for adding the DRIFT-PHOENIX atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_file = os.path.join(input_path, 'drift-phoenix.tgz')
    data_folder = os.path.join(input_path, 'drift-phoenix/')

    url = 'https://people.phys.ethz.ch/~stolkert/species/drift-phoenix.tgz'

    if not os.path.isfile(data_file):
        print('Downloading DRIFT-PHOENIX model spectra (151 MB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking DRIFT-PHOENIX model spectra...', end='', flush=True)
    tar = tarfile.open(data_file)
    tar.extractall(input_path)
    tar.close()
    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    wavelength = None
    flux = []

    for _, _, file_list in os.walk(data_folder):
        for filename in sorted(file_list):

            if filename.startswith('lte_'):
                print_message = f'Adding DRIFT-PHOENIX model spectra... {filename}'
                print(f'\r{print_message:<65}', end='')

                teff.append(float(filename[4:8]))
                logg.append(float(filename[9:12]))
                feh.append(float(filename[12:16]))

                data = np.loadtxt(data_folder+filename)

                if wavelength is None:
                    # [Angstrom] -> [micron]
                    wavelength = data[:, 0]*1e-4

                if np.all(np.diff(wavelength) < 0):
                    raise ValueError('The wavelengths are not all sorted by increasing value.')

                # [erg s-1 cm-2 Angstrom-1] -> [W m-2 micron-1]
                flux.append(data[:, 1]*1e-7*1e4*1e4)

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('drift-phoenix', ('teff', 'logg', 'feh'), database, data_sorted)

    print_message = 'Adding DRIFT-PHOENIX model spectra... [DONE]'
    print(f'\r{print_message:<65}')



"""
Module with reading functionalities for Planck spectra.
"""

import os
import math
import configparser

import numpy as np

from species.analysis import photometry
from species.core import box, constants
from species.read import read_filter


class ReadPlanck:
    """
    Class for reading a Planck spectrum.
    """

    def __init__(self,
                 wavel_range=None,
                 filter_name=None):
        """
        Parameters
        ----------
        wavel_range : tuple(float, float), None
            Wavelength range (micron). A wavelength range of 0.1-1000 micron is used if set to
            None. Not used if ``filter_name`` is not None.
        filter_name : str, None
            Filter name that is used for the wavelength range. The ``wavel_range`` is used if set
            to None.

        Returns
        -------
        NoneType
            None
        """

        self.spectrum_interp = None
        self.wl_points = None
        self.wl_index = None

        self.filter_name = filter_name
        self.wavel_range = wavel_range

        if self.filter_name is not None:
            transmission = read_filter.ReadFilter(self.filter_name)
            self.wavel_range = transmission.wavelength_range()

        elif self.wavel_range is None:
            self.wavel_range = (0.1, 1000.)

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    @staticmethod
    def planck(wavel_points,
               temperature,
               scaling):
        """
        Internal function for calculating a Planck function.

        Parameters
        ----------
        wavel_points : numpy.ndarray
            Wavelength points (micron).
        temperature : float
            Temperature (K).
        scaling : float
            Scaling parameter.

        Returns
        -------
        numpy.ndarray
            Flux density (W m-2 micron-1).
        """

        planck_1 = 2.*constants.PLANCK*constants.LIGHT**2/(1e-6*wavel_points)**5

        planck_2 = np.exp(constants.PLANCK*constants.LIGHT /
                          (1e-6*wavel_points*constants.BOLTZMANN*temperature)) - 1.

        return 1e-6 * 4.*math.pi * scaling * planck_1/planck_2  # [W m-2 micron-1]

    @staticmethod
    def update_parameters(model_param):
        """
        Internal function for updating the dictionary with model parameters.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and 'distance' (pc). The values
            of 'teff' and 'radius' can be a single float, or a list with floats for a combination
            of multiple Planck functions, e.g.
            {'teff': [1500., 1000.], 'radius': [1., 2.], 'distance': 10.}.

        Returns
        -------
        dict
            Updated dictionary with model parameters.
        """

        updated_param = {}

        for i, _ in enumerate(model_param['teff']):
            updated_param[f'teff_{i}'] = model_param['teff'][i]
            updated_param[f'radius_{i}'] = model_param['radius'][i]

        updated_param['distance'] = model_param['distance']

        return updated_param

    def get_spectrum(self,
                     model_param,
                     spec_res):
        """
        Function for calculating a Planck spectrum or a combination of multiple Planck spectra.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and 'distance' (pc). The values
            of 'teff' and 'radius' can be a single float, or a list with floats for a combination
            of multiple Planck functions, e.g.
            {'teff': [1500., 1000.], 'radius': [1., 2.], 'distance': 10.}.
        spec_res : float
            Spectral resolution.

        Returns
        -------
        species.core.box.ModelBox
            Box with the Planck spectrum.
        """

        if 'teff' in model_param and isinstance(model_param['teff'], list):
            model_param = self.update_parameters(model_param)

        wavel_points = [self.wavel_range[0]]

        while wavel_points[-1] <= self.wavel_range[1]:
            wavel_points.append(wavel_points[-1] + wavel_points[-1]/spec_res)

        wavel_points = np.asarray(wavel_points)  # [micron]

        n_planck = (len(model_param)-1) // 2

        if n_planck == 1:
            scaling = ((model_param['radius']*constants.R_JUP) /
                       (model_param['distance']*constants.PARSEC))**2

            flux = self.planck(np.copy(wavel_points),
                               model_param['teff'],
                               scaling)  # [W m-2 micron-1]

        else:
            flux = np.zeros(wavel_points.shape)

            for i in range(n_planck):
                scaling = ((model_param[f'radius_{i}']*constants.R_JUP) /
                           (model_param['distance']*constants.PARSEC))**2

                flux += self.planck(np.copy(wavel_points),
                                    model_param[f'teff_{i}'],
                                    scaling)  # [W m-2 micron-1]

        return box.create_box(boxtype='model',
                              model='planck',
                              wavelength=wavel_points,
                              flux=flux,
                              parameters=model_param,
                              quantity='flux')

    def get_flux(self,
                 model_param,
                 synphot=None):
        """
        Function for calculating the average flux density for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and 'distance' (pc).
        synphot : species.analysis.photometry.SyntheticPhotometry, None
            Synthetic photometry object. The object is created if set to None.

        Returns
        -------
        float
            Average flux density (W m-2 micron-1).
        """

        if 'teff' in model_param and isinstance(model_param['teff'], list):
            model_param = self.update_parameters(model_param)

        spectrum = self.get_spectrum(model_param, 100.)

        if synphot is None:
            synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(spectrum.wavelength, spectrum.flux)



"""
Module with reading functionalities for calibration spectra.
"""

import os
import configparser

import h5py
import spectres
import numpy as np

from scipy.optimize import curve_fit

from species.analysis import photometry
from species.core import box
from species.read import read_filter


class ReadCalibration:
    """
    Class for reading a calibration spectrum from the database.
    """

    def __init__(self,
                 tag,
                 filter_name=None):
        """
        Parameters
        ----------
        tag : str
            Database tag of the calibration spectrum.
        filter_name : str, None
            Filter ID that is used for the wavelength range. Full spectrum is used if set to None.

        Returns
        -------
        NoneType
            None
        """

        self.tag = tag
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = read_filter.ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def resample_spectrum(self,
                          wavel_points,
                          apply_mask=False):
        """
        Function for resampling of a spectrum and uncertainties onto a new wavelength grid.

        Parameters
        ----------
        wavel_points : numpy.ndarray
            Wavelength points (micron).
        apply_mask : bool
            Exclude negative values and NaN values.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the resampled spectrum.
        """

        calibbox = self.get_spectrum()

        if apply_mask:
            indices = np.where(calibbox.flux > 0.)[0]

            calibbox.wavelength = calibbox.wavelength[indices]
            calibbox.flux = calibbox.flux[indices]
            calibbox.error = calibbox.error[indices]

        flux_new, error_new = spectres.spectres(new_spec_wavs=wavel_points,
                                                old_spec_wavs=calibbox.wavelength,
                                                spec_fluxes=calibbox.flux,
                                                spec_errs=calibbox.error)

        return box.create_box(boxtype='spectrum',
                              spectrum='calibration',
                              wavelength=wavel_points,
                              flux=flux_new,
                              error=error_new,
                              name=self.tag,
                              simbad=None,
                              sptype=None,
                              distance=None)

    def get_spectrum(self,
                     model_param=None,
                     apply_mask=False,
                     spec_res=None,
                     extrapolate=False,
                     min_wavelength=None):
        """
        Function for selecting the calibration spectrum.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not used if set to None.
        apply_mask : bool
            Exclude negative values and NaN values.
        spec_res : float, None
            Spectral resolution. Original wavelength points are used if set to None.
        extrapolate : bool
            Extrapolate to 6 micron by fitting a power law function.
        min_wavelength : float, None
            Minimum wavelength used for fitting the power law function. All data is used if set
            to None.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectrum.
        """

        with h5py.File(self.database, 'r') as h5_file:
            data = np.asarray(h5_file['spectra/calibration/'+self.tag])

            wavelength = np.asarray(data[0, ])
            flux = np.asarray(data[1, ])
            error = np.asarray(data[2, ])

        if apply_mask:
            indices = np.where(flux > 0.)[0]

            wavelength = wavelength[indices]
            flux = flux[indices]
            error = error[indices]

        if model_param is not None:
            flux = model_param['scaling']*flux

        if self.wavel_range is None:
            wl_index = np.ones(wavelength.size, dtype=bool)
        else:
            wl_index = (flux > 0.) & (wavelength > self.wavel_range[0]) & \
                       (wavelength < self.wavel_range[1])

        count = np.count_nonzero(wl_index)

        if count > 0:
            index = np.where(wl_index)[0]

            if index[0] > 0:
                wl_index[index[0] - 1] = True

            if index[-1] < len(wl_index)-1:
                wl_index[index[-1] + 1] = True

            wavelength = wavelength[wl_index]
            flux = flux[wl_index]
            error = error[wl_index]

        if extrapolate:
            def _power_law(wavelength, offset, scaling, power_index):
                return offset + scaling*wavelength**power_index

            if min_wavelength:
                indices = np.where(wavelength > min_wavelength)[0]
            else:
                indices = np.arange(0, wavelength.size, 1)

            popt, pcov = curve_fit(f=_power_law,
                                   xdata=wavelength[indices],
                                   ydata=flux[indices],
                                   p0=(0., np.mean(flux[indices]), -1.),
                                   sigma=error[indices])

            sigma = np.sqrt(np.diag(pcov))

            print(f'Fit result for f(x) = a + b*x^c:')
            print(f'a = {popt[0]} +/- {sigma[0]}')
            print(f'b = {popt[1]} +/- {sigma[1]}')
            print(f'c = {popt[2]} +/- {sigma[2]}')

            while wavelength[-1] <= 6.:
                wl_add = wavelength[-1] + wavelength[-1]/1000.

                wavelength = np.append(wavelength, wl_add)
                flux = np.append(flux, _power_law(wl_add, popt[0], popt[1], popt[2]))
                error = np.append(error, 0.)

        if spec_res is not None:
            wavelength_new = [wavelength[0]]

            while wavelength_new[-1] < wavelength[-1]:
                wavelength_new.append(wavelength_new[-1] + wavelength_new[-1]/spec_res)

            wavelength_new = np.asarray(wavelength_new[:-1])

            value_error = True

            while value_error:
                try:
                    flux_new, error_new = spectres.spectres(new_spec_wavs=wavelength_new,
                                                            old_spec_wavs=wavelength,
                                                            spec_fluxes=flux,
                                                            spec_errs=error)

                    value_error = False

                except ValueError:
                    wavelength_new = wavelength_new[1:-1]
                    value_error = True

            wavelength = wavelength_new
            flux = flux_new
            error = error_new

        return box.create_box(boxtype='spectrum',
                              spectrum='calibration',
                              wavelength=wavelength,
                              flux=flux,
                              error=error,
                              name=self.tag,
                              simbad=None,
                              sptype=None,
                              distance=None)

    def get_flux(self,
                 model_param=None):
        """
        Function for calculating the average flux density for the ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not used if set to None.

        Returns
        -------
        float
            Average flux density (W m-2 micron-1).
        """

        specbox = self.get_spectrum(model_param=model_param)

        synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(specbox.wavelength, specbox.flux)

    def get_magnitude(self,
                      model_param=None,
                      distance=None):
        """
        Function for calculating the apparent magnitude for the ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not used if set to None.
        distance : float, None
            Distance to the calibration objects (pc).

        Returns
        -------
        float
            Apparent magnitude (mag).
        float, None
            Absolute magnitude (mag).
        """

        specbox = self.get_spectrum(model_param=model_param)

        synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_magnitude(specbox.wavelength,
                                             specbox.flux,
                                             distance=distance)



"""
Text
"""

import os
import urllib.request

import numpy as np

from astropy.io import fits


def add_vega(input_path, database):
    """
    Function for adding a Vega spectrum to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    data_file = os.path.join(input_path, 'alpha_lyr_stis_008.fits')
    url = 'http://ssb.stsci.edu/cdbs/calspec/alpha_lyr_stis_008.fits'

    if not os.path.isfile(data_file):
        print('Downloading Vega spectrum (270 kB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    if 'spectra/calibration' not in database:
        database.create_group('spectra/calibration')

    if 'spectra/calibration/vega' in database:
        del database['spectra/calibration/vega']

    hdu = fits.open(data_file)
    data = hdu[1].data
    wavelength = data['WAVELENGTH']  # [Angstrom]
    flux = data['FLUX']  # [erg s-1 cm-2 A-1]
    error_stat = data['STATERROR']  # [erg s-1 cm-2 A-1]
    error_sys = data['SYSERROR']  # [erg s-1 cm-2 A-1]
    hdu.close()

    wavelength *= 1e-4  # [Angstrom] -> [micron]
    flux *= 1.e-3*1e4  # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]
    error_stat *= 1.e-3*1e4  # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]
    error_sys *= 1.e-3*1e4  # [erg s-1 cm-2 A-1] -> [W m-2 micron-1]

    print('Adding Vega spectrum...', end='', flush=True)

    database.create_dataset('spectra/calibration/vega',
                            data=np.vstack((wavelength, flux, error_stat)))

    print(' [DONE]')



"""
Module with reading functionalities for spectral libraries.
"""

import os
import configparser

import h5py
import numpy as np

from species.analysis import photometry
from species.core import box
from species.data import database
from species.read import read_filter


class ReadSpectrum:
    """
    Class for reading spectral library data from the database.
    """

    def __init__(self,
                 spec_library,
                 filter_name=None):
        """
        Parameters
        ----------
        spec_library : str
            Name of the spectral library ('irtf', 'spex') or other type of spectrum ('vega').
        filter_name : str, None
            Filter ID for the wavelength range. Full spectra are read if set to None.

        Returns
        -------
        NoneType
            None
        """

        self.spec_library = spec_library
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = read_filter.ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def get_spectrum(self,
                     sptypes=None,
                     exclude_nan=True):
        """
        Function for selecting spectra from the database.

        Parameters
        ----------
        sptypes : list('str', )
            Spectral types to select from a library. The spectral types should be indicated with
            two characters (e.g. 'M5', 'L2', 'T3'). All spectra are selected if set to None.
        exclude_nan : bool
            Exclude wavelength points for which the flux is NaN.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectra.
        """

        h5_file = h5py.File(self.database, 'r')

        try:
            h5_file['spectra/'+self.spec_library]

        except KeyError:
            h5_file.close()
            species_db = database.Database()
            species_db.add_spectrum(self.spec_library, sptypes)
            h5_file = h5py.File(self.database, 'r')

        list_wavelength = []
        list_flux = []
        list_name = []
        list_simbad = []
        list_sptype = []
        list_distance = []

        for item in h5_file['spectra/'+self.spec_library]:
            data = h5_file['spectra/'+self.spec_library+'/'+item]

            wavelength = data[0, :]  # [micron]
            flux = data[1, :]  # [W m-2 micron-1]

            if data.shape[0] == 3:
                error = data[2, :]  # [W m-2 micron-1]

            if exclude_nan:
                indices = np.isnan(flux)
                indices = np.logical_not(indices)
                indices = np.where(indices)[0]

                wavelength = wavelength[indices]
                flux = flux[indices]

                if data.shape[0] == 3:
                    error = error[indices]

            if self.wavel_range is None:
                wl_index = np.arange(0, len(wavelength), 1)

            else:
                wl_index = (flux > 0.) & (wavelength > self.wavel_range[0]) & \
                           (wavelength < self.wavel_range[1])

            count = np.count_nonzero(wl_index)

            if count > 0:
                index = np.where(wl_index)[0]

                if index[0] > 0:
                    wl_index[index[0] - 1] = True

                if index[-1] < len(wl_index)-1:
                    wl_index[index[-1] + 1] = True

                list_wavelength.append(wavelength[wl_index])
                list_flux.append(flux[wl_index])

                attrs = data.attrs
                if 'name' in attrs:
                    list_name.append(data.attrs['name'])

                if 'simbad' in attrs:
                    list_simbad.append(data.attrs['simbad'])

                if 'sptype' in attrs:
                    list_sptype.append(data.attrs['sptype'])

                if 'distance' in attrs:
                    list_distance.append(data.attrs['distance'])

        specbox = box.SpectrumBox()

        specbox.spec_library = self.spec_library
        specbox.wavelength = np.asarray(list_wavelength)
        specbox.flux = np.asarray(list_flux)

        if list_name:
            specbox.name = np.asarray(list_name)

        if list_simbad:
            specbox.simbad = np.asarray(list_simbad)

        if list_sptype:
            specbox.sptype = np.asarray(list_sptype)

        if list_distance:
            specbox.distance = np.asarray(list_distance)

        if sptypes is not None:
            indices = None

            for item in sptypes:
                if indices is None:
                    indices = np.where(np.chararray.startswith(specbox.sptype, item.encode()))[0]

                else:
                    ind_tmp = np.where(np.chararray.startswith(specbox.sptype, item.encode()))[0]
                    indices = np.append(indices, ind_tmp)

            specbox.wavelength = specbox.wavelength[indices]
            specbox.flux = specbox.flux[indices]
            specbox.name = specbox.name[indices]
            specbox.simbad = specbox.simbad[indices]
            specbox.sptype = specbox.sptype[indices]
            specbox.distance = specbox.distance[indices]

        return specbox

    def get_flux(self,
                 sptypes=None):
        """
        Function for calculating the average flux density for the ``filter_name``.

        Parameters
        ----------
        sptypes : list('str', )
            Spectral types to select from a library. The spectral types should be indicated with
            two characters (e.g. 'M5', 'L2', 'T3'). All spectra are selected if set to None.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        specbox = self.get_spectrum(sptypes=sptypes,
                                    exclude_nan=True)

        filter_profile = read_filter.ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()

        synphot = photometry.SyntheticPhotometry(filter_name=self.filter_name)

        phot_wavel = []
        phot_flux = []

        for i, _ in enumerate(specbox.wavelength):
            flux = synphot.spectrum_to_flux(wavelength=specbox.wavelength[i],
                                            flux=specbox.flux[i])

            phot_wavel.append(mean_wavel)
            phot_flux.append(flux)

        return box.create_box(boxtype='photometry',
                              name=specbox.name,
                              sptype=specbox.sptype,
                              wavelength=phot_wavel,
                              flux=phot_flux,
                              app_mag=None,
                              abs_mag=None)

    def get_magnitude(self,
                      sptypes=None):
        """
        Function for calculating the apparent magnitude for the ``filter_name``.

        Parameters
        ----------
        sptypes : list('str', )
            Spectral types to select from a library. The spectral types should be indicated with
            two characters (e.g. 'M5', 'L2', 'T3'). All spectra are selected if set to None.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        specbox = self.get_spectrum(sptypes=sptypes,
                                    exclude_nan=True)

        filter_profile = read_filter.ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()
        wavelength = np.full(specbox.wavelength.shape[0], mean_wavel)

        synphot = photometry.SyntheticPhotometry(filter_name=self.filter_name)

        app_mag, abs_mag = synphot.spectrum_to_magnitude(wavelength=specbox.wavelength,
                                                         flux=specbox.flux,
                                                         distance=specbox.distance)

        return box.create_box(boxtype='photometry',
                              name=specbox.name,
                              sptype=specbox.sptype,
                              wavelength=wavelength,
                              flux=None,
                              app_mag=app_mag,
                              abs_mag=abs_mag)



"""
Module with reading functionalities for atmospheric model spectra.
"""

import os
import math
import warnings
import configparser

import h5py
import spectres
import numpy as np

from scipy.interpolate import RegularGridInterpolator

from species.analysis import photometry
from species.core import box, constants
from species.data import database
from species.read import read_filter, read_calibration
from species.util import read_util


class ReadModel:
    """
    Class for reading a model spectrum from the database.
    """

    def __init__(self,
                 model,
                 wavel_range=None,
                 filter_name=None):
        """
        Parameters
        ----------
        model : str
            Name of the atmospheric model.
        wavel_range : tuple(float, float), None
            Wavelength range (micron). Full spectrum is selected if set to None. Not used if
            ``filter_name`` is not None.
        filter_name : str, None
            Filter ID that is used for the wavelength range. The ``wavel_range`` is used if set
            to None.

        Returns
        -------
        NoneType
            None
        """

        self.model = model

        self.spectrum_interp = None
        self.wl_points = None
        self.wl_index = None

        self.filter_name = filter_name
        self.wavel_range = wavel_range

        if self.filter_name is not None:
            transmission = read_filter.ReadFilter(self.filter_name)
            self.wavel_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def open_database(self):
        """
        Internal function for opening the `species` database.

        Returns
        -------
        h5py._hl.files.File
            The HDF5 database.
        """

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        database_path = config['species']['database']

        h5_file = h5py.File(database_path, 'r')

        try:
            h5_file['models/'+self.model]
        except KeyError:
            raise ValueError(f'The \'{self.model}\' model spectra are not present in the '
                             f'database.')

        return h5_file

    def wavelength_points(self,
                          hdf5_file):
        """
        Internal function for extracting the wavelength points and indices that are used.

        Parameters
        ----------
        hdf5_file : h5py._hl.files.File
            The HDF5 database.

        Returns
        -------
        numpy.ndarray
            Wavelength points (micron).
        numpy.ndarray
            Array with the size of the original wavelength grid. The booleans indicate if a
            wavelength point was used.
        """

        wl_points = np.asarray(hdf5_file[f'models/{self.model}/wavelength'])

        if self.wavel_range is None:
            wl_index = np.ones(wl_points.shape[0], dtype=bool)

        else:
            wl_index = (wl_points > self.wavel_range[0]) & (wl_points < self.wavel_range[1])
            index = np.where(wl_index)[0]

            if index[0]-1 >= 0:
                wl_index[index[0] - 1] = True

            if index[-1]+1 < wl_index.size:
                wl_index[index[-1] + 1] = True

        return wl_points[wl_index], wl_index

    def interpolate_model(self):
        """
        Internal function for linearly interpolating the grid of model spectra.

        Returns
        -------
        NoneType
            None
        """

        h5_file = self.open_database()

        flux = np.asarray(h5_file[f'models/{self.model}/flux'])
        teff = np.asarray(h5_file[f'models/{self.model}/teff'])
        logg = np.asarray(h5_file[f'models/{self.model}/logg'])

        if self.wl_points is None:
            self.wl_points, self.wl_index = self.wavelength_points(h5_file)

        if self.model in ['ames-cond', 'ames-dusty', 'bt-settl']:
            points = (teff, logg)
            flux = flux[:, :, self.wl_index]

        if self.model in ['bt-nextgen', 'drift-phoenix', 'petitcode-cool-clear']:
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])

            points = (teff, logg, feh)
            flux = flux[:, :, :, self.wl_index]

        elif self.model == 'petitcode-cool-cloudy':
            feh = np.asarray(h5_file['models/petitcode-cool-cloudy/feh'])
            fsed = np.asarray(h5_file['models/petitcode-cool-cloudy/fsed'])

            points = (teff, logg, feh, fsed)
            flux = flux[:, :, :, :, self.wl_index]

        elif self.model == 'petitcode-hot-clear':
            feh = np.asarray(h5_file['models/petitcode-hot-clear/feh'])
            co_ratio = np.asarray(h5_file['models/petitcode-hot-clear/co'])

            points = (teff, logg, feh, co_ratio)
            flux = flux[:, :, :, :, self.wl_index]

        elif self.model == 'petitcode-hot-cloudy':
            feh = np.asarray(h5_file['models/petitcode-hot-cloudy/feh'])
            co_ratio = np.asarray(h5_file['models/petitcode-hot-cloudy/co'])
            fsed = np.asarray(h5_file['models/petitcode-hot-cloudy/fsed'])

            flux = flux[:, :, :, :, :, self.wl_index]
            points = (teff, logg, feh, co_ratio, fsed)

        h5_file.close()

        self.spectrum_interp = RegularGridInterpolator(points=points,
                                                       values=flux,
                                                       method='linear',
                                                       bounds_error=False,
                                                       fill_value=np.nan)

    def get_model(self,
                  model_param,
                  spec_res=None,
                  wavel_resample=None,
                  magnitude=False,
                  smooth=False):
        """
        Function for extracting a model spectrum by linearly interpolating the model grid. The
        parameters values should lie within the boundaries of the grid points that are stored
        in the database. The stored grid points can be inspected with
        :func:`~species.read.read_model.ReadModel.get_points`.

        Parameters
        ----------
        model_param : dict
            Model parameters and values. The values should be within the boundaries of the grid.
            The grid boundaries of the available spectra in the database can be obtained with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.
        spec_res : float, None
            Spectral resolution, achieved by smoothing with a Gaussian kernel. The original
            wavelength points are used if both ``spec_res`` and ``wavel_resample`` are set to None.
        wavel_resample : numpy.ndarray
            Wavelength points (micron) to which the spectrum is resampled. Only used if
            ``spec_res`` is set to None.
        magnitude : bool
            Normalize the spectrum with a flux calibrated spectrum of Vega and return the magnitude
            instead of flux density.
        smooth : bool
            If True, the spectral resolution is changed by smoothing with a Gaussian kernel. In
            that case the original wavelength grid is maintained. Note that this requires
            equally-spaced wavelength bins. If False, the spectrum is resampled by linearly
            interpolating the extracted spectrum. Only used if ``spec_res`` is not None.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        if spec_res is not None and wavel_resample is not None:
            raise ValueError('The \'spec_res\' and \'wavel_resample\' parameters can not be used '
                             'simultaneously. Please set one of them to None.')

        if smooth and wavel_resample is not None:
            warnings.warn('The \'smooth\' parameter is ignored because it can only be used in '
                          'combination with \'spec_res\'.')

        grid_bounds = self.get_bounds()

        for key, value in model_param.items():
            if key not in ['radius', 'distance', 'mass', 'luminosity']:
                if key not in self.get_parameters():
                    raise ValueError(f'The \'{key}\' parameter is not required by \'{self.model}\''
                                     f'. The mandatory parameters are {self.get_parameters()}.')

                if value < grid_bounds[key][0]:
                    raise ValueError(f'The input value of \'{key}\' is smaller than the lower '
                                     f'boundary of the model grid ({value} < '
                                     f'{grid_bounds[key][0]}).')

                if value > grid_bounds[key][1]:
                    raise ValueError(f'The input value of \'{key}\' is larger than the upper '
                                     f'boundary of the model grid ({value} > '
                                     f'{grid_bounds[key][1]}).')

        if 'mass' in model_param:
            mass = 1e3 * model_param['mass'] * constants.M_JUP  # [g]
            radius = math.sqrt(1e3 * constants.GRAVITY * mass / (10.**model_param['logg']))  # [cm]
            model_param['radius'] = 1e-2 * radius / constants.R_JUP  # [Rjup]

        if self.spectrum_interp is None:
            self.interpolate_model()

        if self.wavel_range is None:
            wl_points = self.get_wavelengths()
            self.wavel_range = (wl_points[0], wl_points[-1])

        if self.model in ['ames-cond', 'ames-dusty', 'bt-settl']:
            parameters = [model_param['teff'],
                          model_param['logg']]

        elif self.model in ['bt-nextgen', 'drift-phoenix', 'petitcode-cool-clear']:
            parameters = [model_param['teff'],
                          model_param['logg'],
                          model_param['feh']]

        elif self.model == 'petitcode-cool-cloudy':
            parameters = [model_param['teff'],
                          model_param['logg'],
                          model_param['feh'],
                          model_param['fsed']]

        elif self.model == 'petitcode-hot-clear':
            parameters = [model_param['teff'],
                          model_param['logg'],
                          model_param['feh'],
                          model_param['co']]

        elif self.model == 'petitcode-hot-cloudy':
            parameters = [model_param['teff'],
                          model_param['logg'],
                          model_param['feh'],
                          model_param['co'],
                          model_param['fsed']]

        flux = self.spectrum_interp(parameters)[0]

        if 'radius' in model_param:
            model_param['mass'] = read_util.get_mass(model_param)

            if 'distance' in model_param:
                scaling = (model_param['radius']*constants.R_JUP)**2 / \
                          (model_param['distance']*constants.PARSEC)**2

                flux *= scaling

        if spec_res is None and wavel_resample is not None:
            flux = spectres.spectres(new_spec_wavs=wavel_resample,
                                     old_spec_wavs=self.wl_points,
                                     spec_fluxes=flux)

        elif spec_res is not None:
            index = np.where(np.isnan(flux))[0]

            if index.size > 0:
                raise ValueError('Flux values should not contains NaNs. Please make sure that '
                                 'the parameter values and the wavelength range are within '
                                 'the grid boundaries as stored in the database.')

            if smooth:
                flux = read_util.smooth_spectrum(wavelength=self.wl_points,
                                                 flux=flux,
                                                 spec_res=spec_res,
                                                 size=11)

            else:
                wavel_resample = [self.wl_points[0]]
                while wavel_resample[-1] <= self.wl_points[-1]:
                    wavel_resample.append(wavel_resample[-1] + wavel_resample[-1]/spec_res)

                wavel_resample = np.asarray(wavel_resample[:-1])

                indices = np.where((wavel_resample > self.wl_points[0]) &
                                   (wavel_resample < self.wl_points[-2]))[0]

                for i in range(10):
                    try:
                        index_error = False

                        flux = spectres.spectres(new_spec_wavs=wavel_resample[indices][i:-i],
                                                 old_spec_wavs=self.wl_points,
                                                 spec_fluxes=flux)

                    except ValueError:
                        index_error = True

                    if not index_error:
                        wavel_resample = wavel_resample[indices][i:-i]
                        break

        if magnitude:
            quantity = 'magnitude'

            with h5py.File(self.database, 'r') as h5_file:
                try:
                    h5_file['spectra/calibration/vega']

                except KeyError:
                    h5_file.close()
                    species_db = database.Database()
                    species_db.add_spectrum('vega')
                    h5_file = h5py.File(self.database, 'r')

            readcalib = read_calibration.ReadCalibration('vega', filter_name=None)
            calibbox = readcalib.get_spectrum()

            flux_vega, _ = spectres.spectres(new_spec_wavs=self.wl_points,
                                             old_spec_wavs=calibbox.wavelength,
                                             spec_fluxes=calibbox.flux,
                                             spec_errs=calibbox.error)

            flux = -2.5*np.log10(flux/flux_vega)

        else:
            quantity = 'flux'

        is_finite = np.where(np.isfinite(flux))[0]

        if wavel_resample is None:
            wavelength = self.wl_points[is_finite]
        else:
            wavelength = wavel_resample[is_finite]

        return box.create_box(boxtype='model',
                              model=self.model,
                              wavelength=wavelength,
                              flux=flux[is_finite],
                              parameters=model_param,
                              quantity=quantity)

    def get_data(self,
                 model_param):
        """
        Function for selecting a model spectrum (without interpolation) for a set of parameter
        values that coincide with the grid points. The stored grid points can be inspected with
        :func:`~species.read.read_model.ReadModel.get_points`.

        Parameters
        ----------
        model_param : dict
            Model parameters and values. Only discrete values from the original grid are possible.
            Else, the nearest grid values are selected.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        for key in model_param:
            if key not in ['radius', 'distance', 'mass', 'luminosity']:
                if key not in self.get_parameters():
                    raise ValueError(f'The \'{key}\' parameter is not required by \'{self.model}\''
                                     f'. The mandatory parameters are {self.get_parameters()}.')

        h5_file = self.open_database()

        wl_points, wl_index = self.wavelength_points(h5_file)

        flux = np.asarray(h5_file[f'models/{self.model}/flux'])
        teff = np.asarray(h5_file[f'models/{self.model}/teff'])
        logg = np.asarray(h5_file[f'models/{self.model}/logg'])

        if self.model in ['ames-cond', 'ames-dusty', 'bt-settl']:
            teff_index = np.argwhere(teff == model_param['teff'])[0]

            if len(teff_index) == 0:
                raise ValueError('Temperature value not found.')

            teff_index = teff_index[0]
            logg_index = np.argwhere(logg == model_param['logg'])[0]

            if len(logg_index) == 0:
                raise ValueError('Surface gravity value not found.')

            logg_index = logg_index[0]

            flux = flux[teff_index, logg_index, wl_index]

        elif self.model in ['bt-nextgen', 'drift-phoenix', 'petitcode-cool-clear']:
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])

            teff_index = np.argwhere(teff == model_param['teff'])[0]

            if len(teff_index) == 0:
                raise ValueError('Temperature value not found.')

            teff_index = teff_index[0]
            logg_index = np.argwhere(logg == model_param['logg'])[0]

            if len(logg_index) == 0:
                raise ValueError('Surface gravity value not found.')

            logg_index = logg_index[0]
            feh_index = np.argwhere(feh == model_param['feh'])[0]

            if len(feh_index) == 0:
                raise ValueError('Metallicity value not found.')

            feh_index = feh_index[0]

            flux = flux[teff_index, logg_index, feh_index, wl_index]

        elif self.model == 'petitcode-cool-cloudy':
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])
            fsed = np.asarray(h5_file['models/'+self.model+'/fsed'])

            teff_index = np.argwhere(teff == model_param['teff'])[0]

            if len(teff_index) == 0:
                raise ValueError('Temperature value not found.')

            teff_index = teff_index[0]
            logg_index = np.argwhere(logg == model_param['logg'])[0]

            if len(logg_index) == 0:
                raise ValueError('Surface gravity value not found.')

            logg_index = logg_index[0]
            feh_index = np.argwhere(feh == model_param['feh'])[0]

            if len(feh_index) == 0:
                raise ValueError('Metallicity value not found.')

            feh_index = feh_index[0]
            fsed_index = np.argwhere(fsed == model_param['fsed'])[0]

            if len(fsed_index) == 0:
                raise ValueError('f_sed value not found.')

            fsed_index = fsed_index[0]

            flux = flux[teff_index, logg_index, feh_index, fsed_index, wl_index]

        elif self.model == 'petitcode-hot-clear':
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])
            co_ratio = np.asarray(h5_file['models/'+self.model+'/co'])

            teff_index = np.argwhere(teff == model_param['teff'])[0]

            if len(teff_index) == 0:
                raise ValueError('Temperature value not found.')

            teff_index = teff_index[0]
            logg_index = np.argwhere(logg == model_param['logg'])[0]

            if len(logg_index) == 0:
                raise ValueError('Surface gravity value not found.')

            logg_index = logg_index[0]
            feh_index = np.argwhere(feh == model_param['feh'])[0]

            if len(feh_index) == 0:
                raise ValueError('Metallicity value not found.')

            feh_index = feh_index[0]
            co_index = np.argwhere(co_ratio == model_param['co'])[0]

            if len(co_index) == 0:
                raise ValueError('C/O value not found.')

            co_index = co_index[0]

            flux = flux[teff_index, logg_index, feh_index, co_index, wl_index]

        elif self.model == 'petitcode-hot-cloudy':
            feh = np.asarray(h5_file['models/'+self.model+'/feh'])
            co_ratio = np.asarray(h5_file['models/'+self.model+'/co'])
            fsed = np.asarray(h5_file['models/'+self.model+'/fsed'])

            teff_index = np.argwhere(teff == model_param['teff'])[0]

            if len(teff_index) == 0:
                raise ValueError('Temperature value not found.')

            teff_index = teff_index[0]
            logg_index = np.argwhere(logg == model_param['logg'])[0]

            if len(logg_index) == 0:
                raise ValueError('Surface gravity value not found.')

            logg_index = logg_index[0]
            feh_index = np.argwhere(feh == model_param['feh'])[0]

            if len(feh_index) == 0:
                raise ValueError('Metallicity value not found.')

            feh_index = feh_index[0]
            co_index = np.argwhere(co_ratio == model_param['co'])[0]

            if len(co_index) == 0:
                raise ValueError('C/O value not found.')

            co_index = co_index[0]
            fsed_index = np.argwhere(fsed == model_param['fsed'])[0]

            if len(fsed_index) == 0:
                raise ValueError('f_sed value not found.')

            fsed_index = fsed_index[0]

            flux = flux[teff_index, logg_index, feh_index, co_index, fsed_index, wl_index]

        if 'radius' in model_param and 'distance' in model_param:
            scaling = (model_param['radius']*constants.R_JUP)**2 / \
                      (model_param['distance']*constants.PARSEC)**2

            flux *= scaling

        h5_file.close()

        return box.create_box(boxtype='model',
                              model=self.model,
                              wavelength=wl_points,
                              flux=flux,
                              parameters=model_param,
                              quantity='flux')

    def get_flux(self,
                 model_param,
                 synphot=None):
        """
        Function for calculating the average flux density for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Model parameters and values.
        synphot : species.analysis.photometry.SyntheticPhotometry, None
            Synthetic photometry object. The object is created if set to None.

        Returns
        -------
        float
            Average flux density (W m-2 micron-1).
        """

        if self.spectrum_interp is None:
            self.interpolate_model()

        spectrum = self.get_model(model_param, None)

        if synphot is None:
            synphot = photometry.SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(spectrum.wavelength, spectrum.flux)

    def get_magnitude(self,
                      model_param):
        """
        Function for calculating the apparent and absolute magnitudes for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Model parameters and values.

        Returns
        -------
        float
            Apparent magnitude (mag).
        float
            Absolute magnitude (mag).
        """

        if self.spectrum_interp is None:
            self.interpolate_model()

        try:
            spectrum = self.get_model(model_param, None)
        except ValueError:
            warnings.warn(f'The set of model parameters {model_param} is outside the grid range '
                          f'{self.get_bounds()} so returning a NaN.')

            return np.nan, np.nan

        if spectrum.wavelength.size == 0:
            app_mag = np.nan
            abs_mag = np.nan

        else:
            synphot = photometry.SyntheticPhotometry(self.filter_name)

            if 'distance' in model_param:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength, spectrum.flux, distance=model_param['distance'])

            else:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength, spectrum.flux, distance=None)

        return app_mag, abs_mag

    def get_bounds(self):
        """
        Function for extracting the grid boundaries.

        Returns
        -------
        dict
            Boundaries of parameter grid.
        """

        h5_file = self.open_database()

        teff = h5_file['models/'+self.model+'/teff']
        logg = h5_file['models/'+self.model+'/logg']

        if self.model in ['ames-cond', 'ames-dusty', 'bt-settl']:
            bounds = {'teff': (teff[0], teff[-1]),
                      'logg': (logg[0], logg[-1])}

        elif self.model in ['drift-phoenix', 'bt-nextgen', 'petitcode-cool-clear']:
            feh = h5_file['models/'+self.model+'/feh']

            bounds = {'teff': (teff[0], teff[-1]),
                      'logg': (logg[0], logg[-1]),
                      'feh': (feh[0], feh[-1])}

        elif self.model in ['petitcode-cool-cloudy', ]:
            feh = h5_file['models/'+self.model+'/feh']
            fsed = h5_file['models/'+self.model+'/fsed']

            bounds = {'teff': (teff[0], teff[-1]),
                      'logg': (logg[0], logg[-1]),
                      'feh': (feh[0], feh[-1]),
                      'fsed': (fsed[0], fsed[-1])}

        elif self.model in ['petitcode-hot-clear', ]:
            feh = h5_file['models/'+self.model+'/feh']
            co_ratio = h5_file['models/'+self.model+'/co']

            bounds = {'teff': (teff[0], teff[-1]),
                      'logg': (logg[0], logg[-1]),
                      'feh': (feh[0], feh[-1]),
                      'co': (co_ratio[0], co_ratio[-1])}

        elif self.model in ['petitcode-hot-cloudy', ]:
            feh = h5_file['models/'+self.model+'/feh']
            co_ratio = h5_file['models/'+self.model+'/co']
            fsed = h5_file['models/'+self.model+'/fsed']

            bounds = {'teff': (teff[0], teff[-1]),
                      'logg': (logg[0], logg[-1]),
                      'feh': (feh[0], feh[-1]),
                      'co': (co_ratio[0], co_ratio[-1]),
                      'fsed': (fsed[0], fsed[-1])}

        h5_file.close()

        return bounds

    def get_wavelengths(self):
        """
        Function for extracting the wavelength points.

        Returns
        -------
        numpy.ndarray
            Wavelength points (micron).
        """

        h5_file = self.open_database()
        wavelength = np.asarray(h5_file['models/'+self.model+'/wavelength'])
        h5_file.close()

        return wavelength

    def get_points(self):
        """
        Function for extracting the grid points.

        Returns
        -------
        dict
            Parameter points of the model grid.
        """

        points = {}

        h5_file = self.open_database()

        teff = h5_file[f'models/{self.model}/teff']
        logg = h5_file[f'models/{self.model}/logg']

        points['teff'] = np.asarray(teff)
        points['logg'] = np.asarray(logg)

        feh_models = ['drift-phoenix',
                      'bt-nextgen',
                      'petitcode-cool-clear',
                      'petitcode-cool-cloudy',
                      'petitcode-hot-clear',
                      'petitcode-hot-cloudy']

        if self.model in feh_models:
            feh = h5_file[f'models/{self.model}/feh']
            points['feh'] = np.asarray(feh)

        if self.model in ['petitcode-cool-cloudy', 'petitcode-hot-cloudy']:
            fsed = h5_file[f'models/{self.model}/fsed']
            points['fsed'] = np.asarray(fsed)

        if self.model in ['petitcode-hot-clear', 'petitcode-hot-cloudy']:
            co_ratio = h5_file[f'models/{self.model}/co']
            points['co'] = np.asarray(co_ratio)

        h5_file.close()

        return points

    def get_parameters(self):
        """
        Function for extracting the parameter names.

        Returns
        -------
        list(str, )
            Model parameters.
        """

        h5_file = self.open_database()

        dset = h5_file['models/'+self.model]
        nparam = dset.attrs['nparam']

        param = []
        for i in range(nparam):
            param.append(dset.attrs['parameter'+str(i)])

        h5_file.close()

        return param



"""
Utility functions for data processing.
"""

import warnings

import numpy as np


def update_sptype(sptypes):
    """
    Function to update a list with spectral types to two characters (e.g., M8, L3, or T1).

    Parameters
    ----------
    sptypes : numpy.ndarray
        Input spectral types.

    Returns
    -------
    numpy.ndarray
        Updated spectral types.
    """

    sptype_list = ['O', 'B', 'A', 'F', 'G', 'K', 'M', 'L', 'T', 'Y']

    for i, spt_item in enumerate(sptypes):
        if spt_item == 'None':
            pass

        elif spt_item == 'null':
            sptypes[i] = 'None'

        else:
            for list_item in sptype_list:
                try:
                    sp_index = spt_item.index(list_item)
                    sptypes[i] = spt_item[sp_index:sp_index+2]

                except ValueError:
                    pass

    return sptypes


def update_filter(filter_in):
    """
    Function to update a filter ID from the Vizier Photometry viewer VOTable to the filter ID from
    the SVO Filter Profile Service.

    Parameters
    ----------
    filter_in : str
        Filter ID in the format of the Vizier Photometry viewer.

    Returns
    -------
    str
        Filter ID in the format of the SVO Filter Profile Service.
    """

    if filter_in[0:5] == b'2MASS':
        filter_out = str(b'2MASS/2MASS.'+filter_in[6:])

    elif filter_in[0:4] == b'WISE':
        filter_out = str(b'WISE/WISE.'+filter_in[5:])

    elif filter_in[0:10] == b'GAIA/GAIA2':
        filter_out = str(filter_in[0:9]+b'0'+filter_in[10:])

    else:
        filter_out = None

    return filter_out


def sort_data(teff,
              logg,
              feh,
              co,
              fsed,
              wavelength,
              flux):
    """
    Parameters
    ----------
    teff : numpy.ndarray
    logg : numpy.ndarray
    feh : numpy.ndarray, None
    co : numpy.ndarray, None
    fsed : numpy.ndarray, None
    wavelength : numpy.ndarray
    flux : numpy.ndarray

    Returns
    -------
    list(numpy.ndarray, )
    """

    teff_unique = np.unique(teff)
    logg_unique = np.unique(logg)

    if feh is None and co is None and fsed is None:
        spectrum = np.zeros((teff_unique.shape[0],
                             logg_unique.shape[0],
                             wavelength.shape[0]))

    elif feh is not None and co is None and fsed is None:
        feh_unique = np.unique(feh)

        spectrum = np.zeros((teff_unique.shape[0],
                             logg_unique.shape[0],
                             feh_unique.shape[0],
                             wavelength.shape[0]))

    elif feh is not None and co is not None and fsed is None:
        feh_unique = np.unique(feh)
        co_unique = np.unique(co)

        spectrum = np.zeros((teff_unique.shape[0],
                             logg_unique.shape[0],
                             feh_unique.shape[0],
                             co_unique.shape[0],
                             wavelength.shape[0]))

    elif feh is not None and co is None and fsed is not None:
        feh_unique = np.unique(feh)
        fsed_unique = np.unique(fsed)

        spectrum = np.zeros((teff_unique.shape[0],
                             logg_unique.shape[0],
                             feh_unique.shape[0],
                             fsed_unique.shape[0],
                             wavelength.shape[0]))

    else:
        feh_unique = np.unique(feh)
        co_unique = np.unique(co)
        fsed_unique = np.unique(fsed)

        spectrum = np.zeros((teff_unique.shape[0],
                             logg_unique.shape[0],
                             feh_unique.shape[0],
                             co_unique.shape[0],
                             fsed_unique.shape[0],
                             wavelength.shape[0]))

    for i in range(teff.shape[0]):
        index_teff = np.argwhere(teff_unique == teff[i])[0][0]
        index_logg = np.argwhere(logg_unique == logg[i])[0][0]

        if feh is None and co is None and fsed is None:
            spectrum[index_teff, index_logg, :] = flux[i]

        elif feh is not None and co is None and fsed is None:
            index_feh = np.argwhere(feh_unique == feh[i])[0][0]
            spectrum[index_teff, index_logg, index_feh, :] = flux[i]

        elif feh is not None and co is not None and fsed is None:
            index_feh = np.argwhere(feh_unique == feh[i])[0][0]
            index_co = np.argwhere(co_unique == co[i])[0][0]
            spectrum[index_teff, index_logg, index_feh, index_co, :] = flux[i]

            # for j, item in enumerate(flux[i]):
            #     spectrum[index_teff, index_logg, index_feh, index_co, j] = item

        elif feh is not None and co is None and fsed is not None:
            index_feh = np.argwhere(feh_unique == feh[i])[0][0]
            index_fsed = np.argwhere(fsed_unique == fsed[i])[0][0]
            spectrum[index_teff, index_logg, index_feh, index_fsed, :] = flux[i]

        else:
            index_feh = np.argwhere(feh_unique == feh[i])[0][0]
            index_co = np.argwhere(co_unique == co[i])[0][0]
            index_fsed = np.argwhere(fsed_unique == fsed[i])[0][0]
            spectrum[index_teff, index_logg, index_feh, index_co, index_fsed, :] = flux[i]

    sorted_data = [teff_unique, logg_unique]

    if feh is not None:
        sorted_data.append(feh_unique)

    if co is not None:
        sorted_data.append(co_unique)

    if fsed is not None:
        sorted_data.append(fsed_unique)

    sorted_data.append(wavelength)
    sorted_data.append(spectrum)

    return sorted_data


def write_data(model,
               parameters,
               database,
               data_sorted):
    """
    Function for writing the model spectra and parameters to the database.

    Parameters
    ----------
    model : str
        Atmosphere model.
    parameters : list(str, )
        Model parameters.
    database: h5py._hl.files.File
        Database.
    data_sorted : list(numpy.ndarray, )
        Sorted model data with the parameter values, wavelength points (micron), and flux
        densities (W m-2 micron-1).

    Returns
    -------
    NoneType
        None
    """

    if 'models/'+model in database:
        del database[f'models/{model}']

    dset = database.create_group(f'models/{model}')

    dset.attrs['nparam'] = len(parameters)

    for i, item in enumerate(parameters):
        dset.attrs[f'parameter{i}'] = item

        database.create_dataset(f'models/{model}/{item}',
                                data=data_sorted[i])

    database.create_dataset(f'models/{model}/wavelength',
                            data=data_sorted[len(parameters)])

    database.create_dataset(f'models/{model}/flux',
                            data=data_sorted[len(parameters)+1])


def add_missing(model,
                parameters,
                database):
    """
    Parameters
    ----------
    model : str
        Atmosphere model.
    parameters : list(str, )
        Model parameters.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    grid_shape = []
    for item in parameters:
        grid_shape.append(database[f'models/{model}/{item}'].shape[0])

    teff = np.asarray(database[f'models/{model}/teff'])
    flux = np.asarray(database[f'models/{model}/flux'])

    for i in range(grid_shape[0]):
        for j in range(grid_shape[1]):

            if len(parameters) == 2:
                if np.count_nonzero(flux[i, j]) == 0:
                    try:
                        scaling = (teff[i+1]-teff[i])/(teff[i+1]-teff[i-1])
                        flux[i, j] = scaling*flux[i+1, j] + (1.-scaling)*flux[i-1, j]

                    except IndexError:
                        flux[i, j] = np.nan
                        warnings.warn(f'Interpolation is not possible at the edge of the '
                                      f'parameter grid. A NaN value is stored for Teff = '
                                      f'{teff[i]} K.')

            elif len(parameters) == 3:
                for k in range(grid_shape[2]):
                    if np.count_nonzero(flux[i, j, k]) == 0:
                        try:
                            scaling = (teff[i+1]-teff[i])/(teff[i+1]-teff[i-1])
                            flux[i, j, k] = scaling*flux[i+1, j, k] + (1.-scaling)*flux[i-1, j, k]

                        except IndexError:
                            flux[i, j, k] = np.nan
                            warnings.warn(f'Interpolation is not possible at the edge of the '
                                          f'parameter grid. A NaN value is stored for Teff = '
                                          f'{teff[i]} K.')

            elif len(parameters) == 4:
                for k in range(grid_shape[2]):
                    for m in range(grid_shape[3]):
                        if np.count_nonzero(flux[i, j, k, m]) == 0:
                            try:
                                scaling = (teff[i+1]-teff[i])/(teff[i+1]-teff[i-1])
                                flux[i, j, k, m] = scaling*flux[i+1, j, k, m] + \
                                    (1.-scaling)*flux[i-1, j, k, m]

                            except IndexError:
                                flux[i, j, k, m] = np.nan
                                warnings.warn(f'Interpolation is not possible at the edge of the '
                                              f'parameter grid. A NaN value is stored for Teff = '
                                              f'{teff[i]} K.')

            elif len(parameters) == 5:
                for k in range(grid_shape[2]):
                    for m in range(grid_shape[3]):
                        for n in range(grid_shape[4]):
                            if np.count_nonzero(flux[i, j, k, m, n]) == 0:
                                try:
                                    scaling = (teff[i+1]-teff[i])/(teff[i+1]-teff[i-1])
                                    flux[i, j, k, m, n] = scaling*flux[i+1, j, k, m, n] + \
                                        (1.-scaling)*flux[i-1, j, k, mm ]

                                except IndexError:
                                    flux[i, j, k, m, n] = np.nan
                                    warnings.warn(f'Interpolation is not possible at the edge of the '
                                                  f'parameter grid. A NaN value is stored for Teff = '
                                                  f'{teff[i]} K.')

            else:
                raise ValueError('The interpolation of missing data is not yet been implemented '
                                 'for 6 or more parameters.')

    del database[f'models/{model}/flux']

    database.create_dataset(f'models/{model}/flux',
                            data=flux)


def correlation_to_covariance(cor_matrix,
                              spec_sigma):
    """
    Parameters
    ----------
    cor_matrix : numpy.ndarray
        Correlation matrix of the spectrum.
    spec_sigma : numpy.ndarray
        Uncertainties (W m-2 micron-1).

    Returns
    -------
    numpy.ndarrays
        Covariance matrix of the spectrum.
    """

    cov_matrix = np.zeros(cor_matrix.shape)

    for i in range(cor_matrix.shape[0]):
        for j in range(cor_matrix.shape[1]):
            cov_matrix[i, j] = cor_matrix[i, j]*spec_sigma[i]*spec_sigma[j]

            if i == j:
                assert cor_matrix[i, j] == 1.

    return cov_matrix



"""
Module for isochrone data of evolutionary models.
"""

import h5py
import numpy as np

from species.core import constants


def add_baraffe(database,
                tag,
                filename):
    """
    Function for adding the Baraffe et al. isochrone data to the database. Any of the isochrones
    from  https://phoenix.ens-lyon.fr/Grids/ can be used as input.

    https://ui.adsabs.harvard.edu/abs/2003A%26A...402..701B/

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    tag : str
        Tag name in the database.
    filename : str
        Filename with the isochrones data.

    Returns
    -------
    NoneType
        None
    """

    # read in all the data, ignoring empty lines or lines with '---'
    data = []
    with open(filename) as data_file:
        for line in data_file:
            if '---' in line or line == '\n':
                continue
            else:
                data.append(list(filter(None, line.rstrip().split(' '))))

    isochrones = []

    for line in data:
        if '(Gyr)' in line:
            age = line[-1]

        elif 'lg(g)' in line:
            header = ['M/Ms', 'Teff(K)'] + line[1:]

        else:
            line.insert(0, age)
            isochrones.append(line)

    header = np.asarray(header, dtype=bytes)
    isochrones = np.asarray(isochrones, dtype=float)

    isochrones[:, 0] *= 1e3  # [Myr]
    isochrones[:, 1] *= constants.M_SUN/constants.M_JUP  # [Mjup]

    index_sort = np.argsort(isochrones[:, 0])
    isochrones = isochrones[index_sort, :]

    print(f'Adding isochrones: {tag}...', end='', flush=True)

    bytes_type = h5py.special_dtype(vlen=bytes)

    database.create_dataset('isochrones/'+tag+'/filters',
                            data=header[7:],
                            dtype=bytes_type)

    database.create_dataset('isochrones/'+tag+'/magnitudes',
                            data=isochrones[:, 8:])

    dset = database.create_dataset('isochrones/'+tag+'/evolution',
                                   data=isochrones[:, 0:8])

    dset.attrs['model'] = 'baraffe'

    print(' [DONE]')


def add_marleau(database,
                tag,
                filename):
    """
    Function for adding the Marleau et al. isochrone data to the database. The isochrone data can
    be requested from Gabriel Marleau.

    https://ui.adsabs.harvard.edu/abs/2019A%26A...624A..20M/abstract

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    tag : str
        Tag name in the database.
    filename : str
        Filename with the isochrones data.

    Returns
    -------
    NoneType
        None
    """

    # M      age     S_0             L          S(t)            R        Teff
    # (M_J)  (Gyr)   (k_B/baryon)    (L_sol)    (k_B/baryon)    (R_J)    (K)
    mass, age, _, luminosity, _, radius, teff = np.loadtxt(filename, unpack=True)

    age *= 1e3  # [Myr]
    luminosity = np.log10(luminosity)

    mass_cgs = 1e3 * mass * constants.M_JUP  # [g]
    radius_cgs = 1e2 * radius * constants.R_JUP  # [cm]

    logg = np.log10(1e3 * constants.GRAVITY * mass_cgs / radius_cgs**2)

    print(f'Adding isochrones: {tag}...', end='', flush=True)

    isochrones = np.vstack((age, mass, teff, luminosity, logg))
    isochrones = np.transpose(isochrones)

    index_sort = np.argsort(isochrones[:, 0])
    isochrones = isochrones[index_sort, :]

    dset = database.create_dataset('isochrones/'+tag+'/evolution',
                                   data=isochrones)

    dset.attrs['model'] = 'marleau'

    print(' [DONE]')



"""
Module for petitCODE atmospheric model spectra.
"""

import os
import zipfile
import urllib.request

import numpy as np

from species.core import constants
from species.util import data_util


def add_petitcode_cool_clear(input_path,
                             database):
    """
    Function for adding the petitCODE cool clear atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'linder_molliere_grid/clear/specs/')

    url = 'http://mpia.de/~molliere/online_data/linder_molliere_grid.zip'

    data_file = os.path.join(input_path, 'linder_molliere_grid.zip')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE cool clear model spectra (3.7 GB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE cool clear model spectra...', end='', flush=True)

    with zipfile.ZipFile(data_file, 'r') as zip_ref:
        zip_ref.extractall(input_path)

    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    wavelength = None
    flux = []

    for _, _, files in os.walk(data_folder):
        for filename in files:
            print_message = f'Adding petitCODE cool clear model spectra... {filename}'
            print(f'\r{print_message:<87}', end='')

            file_split = filename.split('_')

            teff.append(float(file_split[2]))
            logg.append(float(file_split[4]))
            feh.append(float(file_split[6]))

            data = np.loadtxt(os.path.join(data_folder, filename))

            if wavelength is None:
                # [cm] -> [micron]
                wavelength = data[:, 0]*1e4

            if np.all(np.diff(wavelength) < 0):
                raise ValueError('The wavelengths are not all sorted by increasing value.')

            # [erg s-1 cm-2 Hz-1] -> [W m-2 micron-1]
            flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-cool-clear',
                         ['teff', 'logg', 'feh'],
                         database,
                         data_sorted)

    print_message = 'Adding petitCODE cool clear model spectra... [DONE]'
    print(f'\r{print_message:<87}')


def add_petitcode_cool_cloudy(input_path,
                              database):
    """
    Function for adding the petitCODE cool cloudy atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    data_folder = os.path.join(input_path, 'linder_molliere_grid/cloudy/specs/')

    url = 'http://mpia.de/~molliere/online_data/linder_molliere_grid.zip'

    data_file = os.path.join(input_path, 'linder_molliere_grid.zip')

    if not os.path.isfile(data_file):
        print('Downloading petitCODE cool cloudy model spectra (3.7 GB)...', end='', flush=True)
        urllib.request.urlretrieve(url, data_file)
        print(' [DONE]')

    print('Unpacking petitCODE cool cloudy model spectra...', end='', flush=True)

    with zipfile.ZipFile(data_file, 'r') as zip_ref:
        zip_ref.extractall(input_path)

    print(' [DONE]')

    teff = []
    logg = []
    feh = []
    fsed = []
    wavelength = None
    flux = []

    for _, _, files in os.walk(data_folder):
        for filename in files:
            print_message = f'Adding petitCODE cool cloudy model spectra... {filename}'
            print(f'\r{print_message:<106}', end='')

            file_split = filename.split('_')

            teff.append(float(file_split[2]))
            logg.append(float(file_split[4]))
            feh.append(float(file_split[6]))
            fsed.append(float(file_split[8]))

            data = np.loadtxt(os.path.join(data_folder, filename))

            if wavelength is None:
                # [cm] -> [micron]
                wavelength = data[:, 0]*1e4

            if np.all(np.diff(wavelength) < 0):
                raise ValueError('The wavelengths are not all sorted by increasing value.')

            # [erg s-1 cm-2 Hz-1] -> [W m-2 micron-1]
            flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      None,
                                      np.asarray(fsed),
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-cool-cloudy',
                         ['teff', 'logg', 'feh', 'fsed'],
                         database,
                         data_sorted)

    print_message = 'Adding petitCODE cool cloudy model spectra... [DONE]'
    print(f'\r{print_message:<106}')


def add_petitcode_hot_clear(input_path,
                            database,
                            data_folder):
    """
    Function for adding the petitCODE hot clear atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    data_folder : str
        Path with input data.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    teff = []
    logg = []
    feh = []
    co_ratio = []
    wavelength = None
    flux = []

    for _, _, files in os.walk(data_folder):
        for filename in files:
            print_message = f'Adding petitCODE hot clear model spectra... {filename}'
            print(f'\r{print_message:<100}', end='')

            teff.append(float(filename[9:13]))
            logg.append(float(filename[19:23]))
            feh.append(float(filename[28:32]))
            co_ratio.append(float(filename[36:40]))

            data = np.loadtxt(os.path.join(data_folder, filename))

            if wavelength is None:
                # [cm] -> [micron]
                wavelength = data[:, 0]*1e4

            if np.all(np.diff(wavelength) < 0):
                raise ValueError('The wavelengths are not all sorted by increasing value.')

            # [erg s-1 cm-2 Hz-1] -> [W m-2 micron-1]
            flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      np.asarray(co_ratio),
                                      None,
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-hot-clear',
                         ['teff', 'logg', 'feh', 'co'],
                         database,
                         data_sorted)

    print_message = 'Adding petitCODE hot clear model spectra... [DONE]'
    print(f'\r{print_message:<100}')


def add_petitcode_hot_cloudy(input_path,
                             database,
                             data_folder):
    """
    Function for adding the petitCODE hot cloudy atmospheric models to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    data_folder : str
        Path with input data.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    teff = []
    logg = []
    feh = []
    co_ratio = []
    fsed = []
    wavelength = None
    flux = []

    for _, _, files in os.walk(data_folder):
        for filename in files:
            print_message = f'Adding petitCODE hot cloudy model spectra... {filename}'
            print(f'\r{print_message:<112}', end='')

            teff.append(float(filename[9:13]))
            logg.append(float(filename[19:23]))
            feh.append(float(filename[28:32]))
            co_ratio.append(float(filename[36:40]))
            fsed.append(float(filename[46:50]))

            data = np.loadtxt(os.path.join(data_folder, filename))

            if wavelength is None:
                # [cm] -> [micron]
                wavelength = data[:, 0]*1e4

            if np.all(np.diff(wavelength) < 0):
                raise ValueError('The wavelengths are not all sorted by increasing value.')

            # [erg s-1 cm-2 Hz-1] -> [W m-2 micron-1]
            flux.append(data[:, 1]*1e-9*constants.LIGHT/(wavelength*1e-6)**2)

    data_sorted = data_util.sort_data(np.asarray(teff),
                                      np.asarray(logg),
                                      np.asarray(feh),
                                      np.asarray(co_ratio),
                                      np.asarray(fsed),
                                      wavelength,
                                      np.asarray(flux))

    data_util.write_data('petitcode-hot-cloudy',
                         ['teff', 'logg', 'feh', 'co', 'fsed'],
                         database,
                         data_sorted)

    print_message = 'Adding petitCODE hot cloudy model spectra... [DONE]'
    print(f'\r{print_message:<112}')



"""
Module with reading functionalities of color and magnitude data from photometric and
spectral libraries.
"""

import os
import configparser

import h5py
import numpy as np

from species.core import box
from species.read import read_spectrum
from species.util import phot_util


class ReadColorMagnitude:
    """
    Class for reading color-magnitude data from the database.
    """

    def __init__(self,
                 library,
                 filters_color,
                 filter_mag):
        """
        Parameters
        ----------
        library : str
            Photometric ('vlm-plx' or 'leggett') or spectral ('irtf' or 'spex') library.
        filters_color : tuple(str, str)
            Filter names for the color. For a photometric library, these have to be present in
            the database (typically in the MKO, 2MASS, or WISE system). For a spectral library,
            any filter names can be provided as long as they overlap with the wavelength range
            of the spectra.
        filter_mag : str
            Filter name for the absolute magnitudes (see also description of ``filters_color``).

        Returns
        -------
        NoneType
            None
        """

        self.library = library
        self.filters_color = filters_color
        self.filter_mag = filter_mag

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        with h5py.File(self.database, 'r') as hdf_file:
            if 'photometry' in hdf_file and self.library in hdf_file['photometry']:
                self.lib_type = 'phot_lib'

            elif 'spectra' in hdf_file and self.library in hdf_file['spectra']:
                self.lib_type = 'spec_lib'

            else:
                raise ValueError(f'The \'{self.library}\' library is not present in the database.')

    def get_color_magnitude(self,
                            object_type=None):
        """
        Function for extracting color-magnitude data from the selected library.

        Parameters
        ----------
        object_type : str, None
            Object type for which the colors and magnitudes are extracted. Either field dwarfs
            ('field') or young/low-gravity objects ('young'). All objects are selected if set
            to None.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the colors and magnitudes.
        """

        h5_file = h5py.File(self.database, 'r')

        if self.lib_type == 'phot_lib':
            sptype = np.asarray(h5_file[f'photometry/{self.library}/sptype'])
            distance = np.asarray(h5_file[f'photometry/{self.library}/distance'])  # [pc]
            flag = np.asarray(h5_file[f'photometry/{self.library}/flag'])

            if object_type is None:
                indices = np.arange(0, np.size(sptype), 1)

            elif object_type == 'field':
                indices = np.where(flag == b'null')[0]

            elif object_type == 'young':
                indices = []

                for j, object_flag in enumerate(flag):
                    if b'young' in object_flag:
                        indices.append(j)

                    elif b'lowg' in object_flag:
                        indices.append(j)

                indices = np.array(indices)

            if indices.size > 0:
                mag1 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_color[0]}'])
                mag2 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_color[1]}'])

            color = mag1 - mag2

            if self.filter_mag == self.filters_color[0]:
                mag = phot_util.apparent_to_absolute(mag1, distance)

            elif self.filter_mag == self.filters_color[1]:
                mag = phot_util.apparent_to_absolute(mag2, distance)

            colormag_box = box.create_box(boxtype='colormag',
                                          library=self.library,
                                          object_type=object_type,
                                          filters_color=self.filters_color,
                                          filter_mag=self.filter_mag,
                                          color=color[indices],
                                          magnitude=mag[indices],
                                          sptype=sptype[indices])

        elif self.lib_type == 'spec_lib':
            read_spec_0 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_color[0])

            read_spec_1 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_color[1])

            read_spec_2 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filter_mag)

            phot_box_0 = read_spec_0.get_magnitude(sptypes=None)
            phot_box_1 = read_spec_1.get_magnitude(sptypes=None)
            phot_box_2 = read_spec_2.get_magnitude(sptypes=None)

            colormag_box = box.create_box(boxtype='colormag',
                                          library=self.library,
                                          object_type=object_type,
                                          filters_color=self.filters_color,
                                          filter_mag=self.filter_mag,
                                          color=phot_box_0.app_mag-phot_box_1.app_mag,
                                          magnitude=phot_box_2.abs_mag,
                                          sptype=phot_box_0.sptype)

        h5_file.close()

        return colormag_box


class ReadColorColor:
    """
    Class for reading color-color data from the database.
    """

    def __init__(self,
                 library,
                 filters_colors):
        """
        Parameters
        ----------
        library : str
            Photometric ('vlm-plx' or 'leggett') or spectral ('irtf' or 'spex') library.
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Filter names for the colors. For a photometric library, these have to be present in
            the database (typically in the MKO, 2MASS, or WISE system). For a spectral library,
            any filter names can be provided as long as they overlap with the wavelength range
            of the spectra.

        Returns
        -------
        NoneType
            None
        """

        self.library = library
        self.filters_colors = filters_colors

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

        with h5py.File(self.database, 'r') as hdf_file:
            if 'photometry' in hdf_file and self.library in hdf_file['photometry']:
                self.lib_type = 'phot_lib'

            elif 'spectra' in hdf_file and self.library in hdf_file['spectra']:
                self.lib_type = 'spec_lib'

            else:
                raise ValueError(f'The \'{self.library}\' library is not present in the database.')

    def get_color_color(self,
                        object_type):
        """
        Function for extracting color-color data from the selected library.

        Parameters
        ----------
        object_type : str, None
            Object type for which the colors and magnitudes are extracted. Either field dwarfs
            ('field') or young/low-gravity objects ('young'). All objects are selected if set
            to None.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the colors.
        """

        h5_file = h5py.File(self.database, 'r')

        if self.lib_type == 'phot_lib':
            sptype = np.asarray(h5_file[f'photometry/{self.library}/sptype'])
            flag = np.asarray(h5_file[f'photometry/{self.library}/flag'])

            if object_type is None:
                indices = np.arange(0, np.size(sptype), 1)

            elif object_type == 'field':
                indices = np.where(flag == b'null')[0]

            elif object_type == 'young':
                indices = []

                for j, object_flag in enumerate(flag):
                    if b'young' in object_flag:
                        indices.append(j)

                    elif b'lowg' in object_flag:
                        indices.append(j)

                indices = np.array(indices)

            mag1 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_colors[0][0]}'])
            mag2 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_colors[0][1]}'])
            mag3 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_colors[1][0]}'])
            mag4 = np.asarray(h5_file[f'photometry/{self.library}/{self.filters_colors[1][1]}'])

            color1 = mag1 - mag2
            color2 = mag3 - mag4

            colorbox = box.create_box(boxtype='colorcolor',
                                      library=self.library,
                                      object_type=object_type,
                                      filters=self.filters_colors,
                                      color1=color1[indices],
                                      color2=color2[indices],
                                      sptype=sptype[indices])

        elif self.lib_type == 'spec_lib':
            read_spec_0 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_colors[0][0])

            read_spec_1 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_colors[0][1])

            read_spec_2 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_colors[1][0])

            read_spec_3 = read_spectrum.ReadSpectrum(spec_library=self.library,
                                                     filter_name=self.filters_colors[1][1])

            phot_box_0 = read_spec_0.get_magnitude(sptypes=None)
            phot_box_1 = read_spec_1.get_magnitude(sptypes=None)
            phot_box_2 = read_spec_2.get_magnitude(sptypes=None)
            phot_box_3 = read_spec_3.get_magnitude(sptypes=None)

            colorbox = box.create_box(boxtype='colorcolor',
                                      library=self.library,
                                      object_type=object_type,
                                      filters=self.filters_colors,
                                      color1=phot_box_0.app_mag-phot_box_1.app_mag,
                                      color2=phot_box_2.app_mag-phot_box_3.app_mag,
                                      sptype=phot_box_0.sptype)

        h5_file.close()

        return colorbox



"""
Module with functions for creating color-magnitude and color-color plot.
"""

import os
import math

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from scipy.interpolate import interp1d
from matplotlib.colorbar import Colorbar

from species.core import box
from species.read import read_object
from species.util import plot_util


mpl.rcParams['font.serif'] = ['Bitstream Vera Serif']
mpl.rcParams['font.family'] = 'serif'

plt.rc('axes', edgecolor='black', linewidth=2.5)


def plot_color_magnitude(boxes,
                         objects=None,
                         mass_labels=None,
                         companion_labels=False,
                         field_range=None,
                         label_x='Color [mag]',
                         label_y='Magnitude [mag]',
                         xlim=None,
                         ylim=None,
                         offset=None,
                         legend='upper left',
                         output='color-magnitude.pdf'):
    """
    Function for creating a color-magnitude diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorMagBox, species.core.box.IsochroneBox, )
        Boxes with the color-magnitude and isochrone data from photometric libraries, spectral
        libraries, and/or atmospheric models. The synthetic data have to be created with
        :func:`~species.read.read_isochrone.ReadIsochrone.get_color_magnitude`. These boxes
        contain synthetic colors and magnitudes for a given age and a range of masses.
    objects : tuple(tuple(str, str, str, str), ),
              tuple(tuple(str, str, str, str, str, str, dict, dict), ), None
        Tuple with individual objects. The objects require a tuple with their database tag, the two
        filter names for the color, and the filter names for the absolute magnitude. Optionally, a
        dictionary with keyword arguments can be provided for the object's marker and label,
        respectively. For example, ``{'marker': 'o', 'ms': 10}`` for the marker and
        ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})`` for the label. Not used if set to
        None.
    mass_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with masses next to the isochrone data of `models`. The list with masses has
        to be provided in Jupiter mass. Alternatively, a list of tuples can be provided with
        the planet mass and position of the label ('left' or 'right), for example
        ``[(10., 'left'), (20., 'right')]``. No labels are shown if set to None.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    field_range : tuple(str, str), None
        Range of the discrete colorbar for the field dwarfs. The tuple should contain the lower
        and upper value ('early M', 'late M', 'early L', 'late L', 'early T', 'late T', 'early Y).
        The full range is used if set to None.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    xlim : tuple(float, float), None
        Limits for the x-axis. Not used if set to None.
    ylim : tuple(float, float), None
        Limits for the y-axis. Not used if set to None.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str, None
        Legend position. Not shown if set to None.
    output : str
        Output filename.

    Returns
    -------
    NoneType
        None

    """

    print(f'Plotting color-magnitude diagram: {output}... ', end='', flush=True)

    model_color = ('#234398', '#f6a432', 'black')
    model_linestyle = ('-', '--', ':', '-.')

    isochrones = []
    models = []
    empirical = []

    for item in boxes:
        if isinstance(item, box.IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, box.ColorMagBox):
            if item.object_type == 'model':
                models.append(item)
            else:
                empirical.append(item)

        else:
            raise ValueError(f'Found a {type(item)} while only ColorMagBox and IsochroneBox '
                             f'objects can be provided to \'boxes\'.')

    if empirical:
        plt.figure(1, figsize=(4., 4.8))
        gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.5])
        gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[2, 0])
        ax2 = plt.subplot(gridsp[0, 0])

    else:
        plt.figure(1, figsize=(4., 4.5))
        gridsp = mpl.gridspec.GridSpec(1, 1)
        gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

        ax1 = plt.subplot(gridsp[0, 0])

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=1, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=1, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset is not None:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)

    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim is not None:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim is not None:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        cmap_teff = plt.cm.afmhot

        teff_min = np.inf
        teff_max = -np.inf

        for item in models:

            if np.amin(item.sptype) < teff_min:
                teff_min = np.amin(item.sptype)

            if np.amax(item.sptype) > teff_max:
                teff_max = np.amax(item.sptype)

        norm_teff = mpl.colors.Normalize(vmin=teff_min, vmax=teff_max)

        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library not in model_dict:
                model_dict[item.library] = [count, 0]
                count += 1

            else:
                model_dict[item.library] = [model_dict[item.library][0],
                                            model_dict[item.library][1]+1]

            model_count = model_dict[item.library]

            if model_count[0] == 3:
                raise ValueError('Only three different types of model atmospheres can be added.')

            if model_count[1] == 0:
                label = plot_util.model_name(item.library)

                ax1.plot(item.color, item.magnitude, linestyle=model_linestyle[model_count[1]],
                         linewidth=1.2, color=model_color[model_count[0]], label=label, zorder=0)

                if mass_labels is not None:
                    interp_magnitude = interp1d(item.sptype, item.magnitude)
                    interp_color = interp1d(item.sptype, item.color)

                    for i, mass_item in enumerate(mass_labels):
                        if isinstance(mass_item, tuple):
                            mass_val = mass_item[0]
                            mass_pos = mass_item[1]

                        else:
                            mass_val = mass_item
                            mass_pos = 'right'

                        if j == 0 or (j > 0 and mass_val < 20.):
                            pos_color = interp_color(mass_val)
                            pos_mag = interp_magnitude(mass_val)

                            if mass_pos == 'left':
                                mass_ha = 'right'
                                mass_xytext = (pos_color-0.05, pos_mag)

                            else:
                                mass_ha = 'left'
                                mass_xytext = (pos_color+0.05, pos_mag)

                            mass_label = str(int(mass_val))+r' M$_\mathregular{J}$'

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if xlim[0]+0.2 < pos_color < xlim[1]-0.2 and \
                                    ylim[1]+0.2 < pos_mag < ylim[0]-0.2:

                                ax1.scatter(pos_color, pos_mag, c=model_color[model_count[0]], s=15,
                                            edgecolor='none', zorder=0)

                                ax1.annotate(mass_label, (pos_color, pos_mag),
                                             color=model_color[model_count[0]], fontsize=9,
                                             xytext=mass_xytext, zorder=1, ha=mass_ha, va='center')

            else:
                ax1.plot(item.color, item.magnitude, linestyle=model_linestyle[model_count[1]],
                         linewidth=0.6, color=model_color[model_count[0]], zorder=0)

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = plot_util.field_bounds_ticks(field_range)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            color = item.color
            magnitude = item.magnitude

            indices = np.where(sptype != b'None')[0]

            sptype = sptype[indices]
            color = color[indices]
            magnitude = magnitude[indices]

            spt_disc = plot_util.sptype_substellar(sptype, color.shape)

            _, unique = np.unique(color, return_index=True)

            sptype = sptype[unique]
            color = color[unique]
            magnitude = magnitude[unique]
            spt_disc = spt_disc[unique]

            if item.object_type == 'field' or item.object_type is None:
                scat = ax1.scatter(color, magnitude, c=spt_disc, cmap=cmap, norm=norm, s=50,
                                   alpha=0.7, edgecolor='none', zorder=2)

                cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                              ticklocation='top', format='%.2f')

                cb.ax.tick_params(width=1, length=5, labelsize=10, direction='in', color='black')

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)

            elif item.object_type == 'young':
                ax1.plot(color, magnitude, marker='s', ms=4, linestyle='none', alpha=0.7,
                         color='gray', markeredgecolor='black', label='Young/low-gravity', zorder=2)

    if isochrones:
        for item in isochrones:
            ax1.plot(item.color, item.magnitude, linestyle='-', linewidth=1.2, color='black')

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = read_object.ReadObject(item[0])

            objcolor1 = objdata.get_photometry(item[1])
            objcolor2 = objdata.get_photometry(item[2])
            abs_mag = objdata.get_absmag(item[3])

            colorerr = math.sqrt(objcolor1[1]**2+objcolor2[1]**2)

            x_color = objcolor1[0]-objcolor2[0]
            y_mag = abs_mag[0]

            if len(item) > 4 and item[4] is not None:
                kwargs = item[4]

            else:
                kwargs = {'marker': '>',
                          'ms': 6.,
                          'color': 'black',
                          'mfc': 'white',
                          'mec': 'black',
                          'label': 'Directly imaged'}

            ax1.errorbar(x_color, y_mag, yerr=abs_mag[1], xerr=colorerr, zorder=3, **kwargs)

            if companion_labels:
                x_range = ax1.get_xlim()
                y_range = ax1.get_ylim()

                if len(item) > 4:
                    kwargs = item[5]

                else:
                    kwargs = {'ha': 'left',
                              'va': 'bottom',
                              'fontsize': 8.5,
                              'xytext': (5., 5.),
                              'color': 'black'}

                ax1.annotate(objdata.object_name, (x_color, y_mag), zorder=3,
                             textcoords='offset points', **kwargs)

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            ax1.legend(by_label.values(), by_label.keys(), loc=legend, fontsize=8.5,
                       frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print('[DONE]')


def plot_color_color(boxes,
                     objects=None,
                     mass_labels=None,
                     companion_labels=False,
                     field_range=None,
                     label_x='color [mag]',
                     label_y='color [mag]',
                     xlim=None,
                     ylim=None,
                     offset=None,
                     legend='upper left',
                     output='color-color.pdf'):
    """
    Function for creating a color-color diagram.

    Parameters
    ----------
    boxes : list(species.core.box.ColorColorBox, species.core.box.IsochroneBox, )
        Boxes with the color-color and isochrone data from photometric libraries, spectral
        libraries, and/or atmospheric models. The synthetic data have to be created with
        :func:`~species.read.read_isochrone.ReadIsochrone.get_color_color`. These boxes
        contain synthetic colors for a given age and a range of masses.
    objects : tuple(tuple(str, str, str, str), ),
              tuple(tuple(str, str, str, str, str, str, dict, dict), ), None
        Tuple with individual objects. The objects require a tuple with their database tag, the two
        filter names for the first color, and the two filter names for the second color.
        Optionally, a dictionary with keyword arguments can be provided for the object's marker and
        label, respectively. For example, ``{'marker': 'o', 'ms': 10}`` for the marker and
        ``{'ha': 'left', 'va': 'bottom', 'xytext': (5, 5)})`` for the label. Not used if set to
        None.
    mass_labels : list(float, ), list(tuple(float, str), ), None
        Plot labels with masses next to the isochrone data of `models`. The list with masses has
        to be provided in Jupiter mass. Alternatively, a list of tuples can be provided with
        the planet mass and position of the label ('left' or 'right), for example
        ``[(10., 'left'), (20., 'right')]``. No labels are shown if set to None.
    companion_labels : bool
        Plot labels with the names of the directly imaged companions.
    field_range : tuple(str, str), None
        Range of the discrete colorbar for the field dwarfs. The tuple should contain the lower
        and upper value ('early M', 'late M', 'early L', 'late L', 'early T', 'late T', 'early Y).
        The full range is used if set to None.
    label_x : str
        Label for the x-axis.
    label_y : str
        Label for the y-axis.
    output : str
        Output filename.
    xlim : tuple(float, float)
        Limits for the x-axis.
    ylim : tuple(float, float)
        Limits for the y-axis.
    offset : tuple(float, float), None
        Offset of the x- and y-axis label.
    legend : str
        Legend position.

    Returns
    -------
    NoneType
        None
    """

    print(f'Plotting color-color diagram: {output}... ', end='', flush=True)

    model_color = ('#234398', '#f6a432')
    model_linestyle = ('-', '--', ':', '-.')

    isochrones = []
    models = []
    empirical = []

    for item in boxes:
        if isinstance(item, box.IsochroneBox):
            isochrones.append(item)

        elif isinstance(item, box.ColorColorBox):
            if item.object_type == 'model':
                models.append(item)
            else:
                empirical.append(item)

        else:
            raise ValueError(f'Found a {type(item)} while only ColorColorBox and IsochroneBox '
                             f'objects can be provided to \'boxes\'.')

    plt.figure(1, figsize=(4, 4.3))
    gridsp = mpl.gridspec.GridSpec(3, 1, height_ratios=[0.2, 0.1, 4.])
    gridsp.update(wspace=0., hspace=0., left=0, right=1, bottom=0, top=1)

    ax1 = plt.subplot(gridsp[2, 0])
    ax2 = plt.subplot(gridsp[0, 0])

    ax1.tick_params(axis='both', which='major', colors='black', labelcolor='black',
                    direction='in', width=1, length=5, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.tick_params(axis='both', which='minor', colors='black', labelcolor='black',
                    direction='in', width=1, length=3, labelsize=12, top=True,
                    bottom=True, left=True, right=True)

    ax1.set_xlabel(label_x, fontsize=14)
    ax1.set_ylabel(label_y, fontsize=14)

    ax1.invert_yaxis()

    if offset:
        ax1.get_xaxis().set_label_coords(0.5, offset[0])
        ax1.get_yaxis().set_label_coords(offset[1], 0.5)
    else:
        ax1.get_xaxis().set_label_coords(0.5, -0.08)
        ax1.get_yaxis().set_label_coords(-0.12, 0.5)

    if xlim:
        ax1.set_xlim(xlim[0], xlim[1])

    if ylim:
        ax1.set_ylim(ylim[0], ylim[1])

    if models is not None:
        cmap_teff = plt.cm.afmhot

        teff_min = np.inf
        teff_max = -np.inf

        for item in models:

            if np.amin(item.sptype) < teff_min:
                teff_min = np.amin(item.sptype)

            if np.amax(item.sptype) > teff_max:
                teff_max = np.amax(item.sptype)

        norm_teff = mpl.colors.Normalize(vmin=teff_min, vmax=teff_max)

        count = 0

        model_dict = {}

        for j, item in enumerate(models):
            if item.library not in model_dict:
                model_dict[item.library] = [count, 0]
                count += 1

            else:
                model_dict[item.library] = [model_dict[item.library][0],
                                            model_dict[item.library][1]+1]

            model_count = model_dict[item.library]

            if model_count[1] == 0:
                label = plot_util.model_name(item.library)

                ax1.plot(item.color1, item.color2, linestyle=model_linestyle[model_count[1]],
                         linewidth=0.6, color=model_color[model_count[0]], label=label, zorder=0)

                if mass_labels is not None:
                    interp_color1 = interp1d(item.sptype, item.color1)
                    interp_color2 = interp1d(item.sptype, item.color2)

                    for i, mass_item in enumerate(mass_labels):
                        if isinstance(mass_item, tuple):
                            mass_val = mass_item[0]
                            mass_pos = mass_item[1]

                        else:
                            mass_val = mass_item
                            mass_pos = 'right'

                        if j == 0 or (j > 0 and mass_val < 20.):
                            pos_color1 = interp_color1(mass_val)
                            pos_color2 = interp_color2(mass_val)

                            if mass_pos == 'left':
                                mass_ha = 'right'
                                mass_xytext = (pos_color1-0.05, pos_color2)

                            else:
                                mass_ha = 'left'
                                mass_xytext = (pos_color1+0.05, pos_color2)

                            mass_label = str(int(mass_val))+r' M$_\mathregular{J}$'

                            xlim = ax1.get_xlim()
                            ylim = ax1.get_ylim()

                            if xlim[0]+0.2 < pos_color1 < xlim[1]-0.2 and \
                                    ylim[0]+0.2 < pos_color2 < ylim[1]-0.2:

                                ax1.scatter(pos_color1, pos_color2, c=model_color[model_count[0]],
                                            s=15, edgecolor='none', zorder=0)

                                ax1.annotate(mass_label, (pos_color1, pos_color2),
                                             color=model_color[model_count[0]], fontsize=9,
                                             xytext=mass_xytext, ha=mass_ha, va='center', zorder=1)

            else:
                ax1.plot(item.color1, item.color2, linestyle=model_linestyle[model_count[1]],
                         linewidth=0.6, color=model_color[model_count[0]], zorder=0)

    if empirical:
        cmap = plt.cm.viridis

        bounds, ticks, ticklabels = plot_util.field_bounds_ticks(field_range)
        norm = mpl.colors.BoundaryNorm(bounds, cmap.N)

        for item in empirical:
            sptype = item.sptype
            color1 = item.color1
            color2 = item.color2

            indices = np.where(sptype != 'None')[0]

            sptype = sptype[indices]
            color1 = color1[indices]
            color2 = color2[indices]

            spt_disc = plot_util.sptype_substellar(sptype, color1.shape)
            _, unique = np.unique(color1, return_index=True)

            sptype = sptype[unique]
            color1 = color1[unique]
            color2 = color2[unique]
            spt_disc = spt_disc[unique]

            if item.object_type == 'field':
                scat = ax1.scatter(color1, color2, c=spt_disc, cmap=cmap, norm=norm, s=50,
                                   alpha=0.7, edgecolor='none', zorder=2)

                cb = Colorbar(ax=ax2, mappable=scat, orientation='horizontal',
                              ticklocation='top', format='%.2f')

                cb.ax.tick_params(width=1, length=5, labelsize=10, direction='in', color='black')

                cb.set_ticks(ticks)
                cb.set_ticklabels(ticklabels)

            elif item.object_type == 'young':
                ax1.plot(color1, color2, marker='s', ms=4, linestyle='none', alpha=0.7,
                         color='gray', markeredgecolor='black', label='Young/low-gravity', zorder=2)

    if isochrones:
        for item in isochrones:
            ax1.plot(item.colors[0], item.colors[1], linestyle='-', linewidth=1.2, color='black')

    if objects is not None:
        for i, item in enumerate(objects):
            objdata = read_object.ReadObject(item[0])

            mag1 = objdata.get_photometry(item[1][0])[0]
            mag2 = objdata.get_photometry(item[1][1])[0]
            mag3 = objdata.get_photometry(item[2][0])[0]
            mag4 = objdata.get_photometry(item[2][1])[0]

            err1 = objdata.get_photometry(item[1][0])[1]
            err2 = objdata.get_photometry(item[1][1])[1]
            err3 = objdata.get_photometry(item[2][0])[1]
            err4 = objdata.get_photometry(item[2][1])[1]

            color1 = mag1 - mag2
            color2 = mag3 - mag4

            error1 = math.sqrt(err1**2+err2**2)
            error2 = math.sqrt(err3**2+err4**2)

            if len(item) > 3 and item[3] is not None:
                kwargs = item[3]

            else:
                kwargs = {'marker': '>',
                          'ms': 6.,
                          'color': 'black',
                          'mfc': 'white',
                          'mec': 'black',
                          'label': 'Directly imaged'}

            ax1.errorbar(color1, color2, xerr=error1, yerr=error2, zorder=3, **kwargs)

            if companion_labels:
                x_range = ax1.get_xlim()
                y_range = ax1.get_ylim()

                if len(item) > 3:
                    kwargs = item[4]

                else:
                    kwargs = {'ha': 'left',
                              'va': 'bottom',
                              'fontsize': 8.5,
                              'xytext': (5., 5.),
                              'color': 'black'}

                ax1.annotate(objdata.object_name, (color1, color2), zorder=3,
                             textcoords='offset points', **kwargs)

    handles, labels = ax1.get_legend_handles_labels()

    if legend is not None:
        handles, labels = ax1.get_legend_handles_labels()

        # prevent duplicates
        by_label = dict(zip(labels, handles))

        if handles:
            ax1.legend(by_label.values(), by_label.keys(), loc=legend, fontsize=8.5,
                       frameon=False, numpoints=1)

    plt.savefig(os.getcwd()+'/'+output, bbox_inches='tight')
    plt.clf()
    plt.close()

    print('[DONE]')



"""
Module with reading functionalities for isochrones.
"""

import os
import configparser

import h5py
import numpy as np

from scipy.interpolate import griddata

from species.core import box
from species.read import read_model


class ReadIsochrone:
    """
    Class for reading an isochrone from the database.
    """

    def __init__(self,
                 tag):
        """
        Parameters
        ----------
        tag : str
            Database tag of the isochrone data.

        Returns
        -------
        NoneType
            None
        """

        self.tag = tag

        config_file = os.path.join(os.getcwd(), 'species_config.ini')

        config = configparser.ConfigParser()
        config.read_file(open(config_file))

        self.database = config['species']['database']

    def get_isochrone(self,
                      age,
                      masses,
                      filters_color,
                      filter_mag):
        """
        Function for selecting an isochrone.

        Parameters
        ----------
        age : str
            Age (Myr) at which the isochrone data is interpolated.
        masses : numpy.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        filters_color : tuple(str, str), None
            Filter IDs for the color as listed in the file with the isochrone data. Not selected if
            set to None or if only evolutionary tracks are available.
        filter_mag : str, None
            Filter ID for the absolute magnitude as listed in the file with the isochrone data. Not
            selected if set to None or if only evolutionary tracks are available.

        Returns
        -------
        species.core.box.IsochroneBox
            Box with the isochrone.
        """

        age_points = np.repeat(age, masses.shape[0])  # [Myr]

        color = None
        mag_abs = None

        index_teff = 2
        index_logg = 4

        with h5py.File(self.database, 'r') as h5_file:
            model = h5_file['isochrones/'+self.tag+'/evolution'].attrs['model']
            evolution = np.asarray(h5_file['isochrones/'+self.tag+'/evolution'])

            if model == 'baraffe':
                filters = list(h5_file['isochrones/'+self.tag+'/filters'])
                magnitudes = np.asarray(h5_file['isochrones/'+self.tag+'/magnitudes'])

        if model == 'baraffe':
            for i, item in enumerate(filters):
                filters[i] = item.decode()

            if filters_color is not None:
                index_color_1 = filters.index(filters_color[0])
                index_color_2 = filters.index(filters_color[1])

            if filter_mag is not None:
                index_mag = filters.index(filter_mag)

            if filters_color is not None:
                mag_color_1 = griddata(points=evolution[:, 0:2],
                                       values=magnitudes[:, index_color_1],
                                       xi=np.stack((age_points, masses), axis=1),
                                       method='linear',
                                       fill_value='nan',
                                       rescale=False)

                mag_color_2 = griddata(points=evolution[:, 0:2],
                                       values=magnitudes[:, index_color_2],
                                       xi=np.stack((age_points, masses), axis=1),
                                       method='linear',
                                       fill_value='nan',
                                       rescale=False)

                color = mag_color_1-mag_color_2

            if filter_mag is not None:
                mag_abs = griddata(points=evolution[:, 0:2],
                                   values=magnitudes[:, index_mag],
                                   xi=np.stack((age_points, masses), axis=1),
                                   method='linear',
                                   fill_value='nan',
                                   rescale=False)

        teff = griddata(points=evolution[:, 0:2],
                        values=evolution[:, index_teff],
                        xi=np.stack((age_points, masses), axis=1),
                        method='linear',
                        fill_value='nan',
                        rescale=False)

        logg = griddata(points=evolution[:, 0:2],
                        values=evolution[:, index_logg],
                        xi=np.stack((age_points, masses), axis=1),
                        method='linear',
                        fill_value='nan',
                        rescale=False)

        return box.create_box(boxtype='isochrone',
                              model=self.tag,
                              filters_color=filters_color,
                              filter_mag=filter_mag,
                              color=color,
                              magnitude=mag_abs,
                              teff=teff,
                              logg=logg,
                              masses=masses)

    def get_color_magnitude(self,
                            age,
                            masses,
                            model,
                            filters_color,
                            filter_mag):
        """
        Function for calculating color-magnitude combinations from a selected isochrone.

        Parameters
        ----------
        age : str
            Age (Myr) at which the isochrone data is interpolated.
        masses : numpy.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        model : str
            Atmospheric model used to compute the synthetic photometry.
        filters_color : tuple(str, str)
            Filter IDs for the color as listed in the file with the isochrone data. The filter IDs
            should be provided in the format of the SVO Filter Profile Service.
        filter_mag : str
            Filter ID for the absolute magnitude as listed in the file with the isochrone data. The
            value should be equal to one of the ``filters_color`` values.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the color-magnitude data.
        """

        isochrone = self.get_isochrone(age=age,
                                       masses=masses,
                                       filters_color=None,
                                       filter_mag=None)

        model1 = read_model.ReadModel(model=model, filter_name=filters_color[0])
        model2 = read_model.ReadModel(model=model, filter_name=filters_color[1])

        if model1.get_parameters() != ['teff', 'logg']:
            raise ValueError('Creating synthetic colors and magnitudes from isochrones is '
                             'currently only implemented for models with only Teff and log(g) '
                             'as free parameters. Please contact Tomas Stolker if additional '
                             'functionalities are required.')

        mag1 = np.zeros(isochrone.masses.shape[0])
        mag2 = np.zeros(isochrone.masses.shape[0])

        for i, mass_item in enumerate(isochrone.masses):
            model_param = {'teff': isochrone.teff[i],
                           'logg': isochrone.logg[i],
                           'mass': mass_item,
                           'distance': 10.}

            if np.isnan(isochrone.teff[i]):
                mag1[i] = np.nan
                mag2[i] = np.nan

            else:
                for item_bounds in model1.get_bounds():
                    if model_param[item_bounds] <= model1.get_bounds()[item_bounds][0]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                    elif model_param[item_bounds] >= model1.get_bounds()[item_bounds][1]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                if not np.isnan(mag1[i]):
                    mag1[i], _ = model1.get_magnitude(model_param=model_param)
                    mag2[i], _ = model2.get_magnitude(model_param=model_param)

        if filter_mag == filters_color[0]:
            abs_mag = mag1

        elif filter_mag == filters_color[1]:
            abs_mag = mag2

        else:
            raise ValueError('The filter_mag argument should be equal to one of the two filter '
                             'values of filters_color.')

        return box.create_box(boxtype='colormag',
                              library=model,
                              object_type='model',
                              filters_color=filters_color,
                              filter_mag=filter_mag,
                              color=mag1-mag2,
                              magnitude=abs_mag,
                              sptype=masses)

    def get_color_color(self,
                        age,
                        masses,
                        model,
                        filters):
        """
        Function for calculating color-magnitude combinations from a selected isochrone.

        Parameters
        ----------
        age : str
            Age (Myr) at which the isochrone data is interpolated.
        masses : numpy.ndarray
            Masses (Mjup) at which the isochrone data is interpolated.
        model : str
            Atmospheric model used to compute the synthetic photometry.
        filters : tuple(tuple(str, str), tuple(str, str))
            Filter IDs for the colors as listed in the file with the isochrone data. The filter IDs
            should be provided in the format of the SVO Filter Profile Service.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the color-color data.
        """

        isochrone = self.get_isochrone(age=age,
                                       masses=masses,
                                       filters_color=None,
                                       filter_mag=None)

        model1 = read_model.ReadModel(model=model, filter_name=filters[0][0])
        model2 = read_model.ReadModel(model=model, filter_name=filters[0][1])
        model3 = read_model.ReadModel(model=model, filter_name=filters[1][0])
        model4 = read_model.ReadModel(model=model, filter_name=filters[1][1])

        if model1.get_parameters() != ['teff', 'logg']:
            raise ValueError('Creating synthetic colors and magnitudes from isochrones is '
                             'currently only implemented for models with only Teff and log(g) '
                             'as free parameters. Please contact Tomas Stolker if additional '
                             'functionalities are required.')

        mag1 = np.zeros(isochrone.masses.shape[0])
        mag2 = np.zeros(isochrone.masses.shape[0])
        mag3 = np.zeros(isochrone.masses.shape[0])
        mag4 = np.zeros(isochrone.masses.shape[0])

        for i, mass_item in enumerate(isochrone.masses):
            model_param = {'teff': isochrone.teff[i],
                           'logg': isochrone.logg[i],
                           'mass': mass_item,
                           'distance': 10.}

            if np.isnan(isochrone.teff[i]):
                mag1[i] = np.nan
                mag2[i] = np.nan

            else:
                for item_bounds in model1.get_bounds():
                    if model_param[item_bounds] <= model1.get_bounds()[item_bounds][0]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                    elif model_param[item_bounds] >= model1.get_bounds()[item_bounds][1]:
                        mag1[i] = np.nan
                        mag2[i] = np.nan

                if not np.isnan(mag1[i]):
                    mag1[i], _ = model1.get_magnitude(model_param=model_param)
                    mag2[i], _ = model2.get_magnitude(model_param=model_param)
                    mag3[i], _ = model3.get_magnitude(model_param=model_param)
                    mag4[i], _ = model4.get_magnitude(model_param=model_param)

        return box.create_box(boxtype='colorcolor',
                              library=model,
                              object_type='model',
                              filters=filters,
                              color1=mag1-mag2,
                              color2=mag3-mag4,
                              sptype=masses)



"""
Module for adding IRTF spectra tot the database.
"""

import os
import tarfile
import urllib.request

import numpy as np

from astropy.io import fits

from species.data import queries
from species.util import data_util


def add_irtf(input_path,
             database,
             sptypes):
    """
    Function to add the IRTF Spectral Library to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        Database.
    sptypes : tuple(str, )
        Spectral types ('F', 'G', 'K', 'M', 'L', 'T').

    Returns
    -------
    NoneType
        None
    """

    datadir = os.path.join(input_path, 'irtf')

    if not os.path.exists(datadir):
        os.makedirs(datadir)

    data_file = {'F': os.path.join(input_path, 'irtf/F_fits_091201.tar'),
                 'G': os.path.join(input_path, 'irtf/G_fits_091201.tar'),
                 'K': os.path.join(input_path, 'irtf/K_fits_091201.tar'),
                 'M': os.path.join(input_path, 'irtf/M_fits_091201.tar'),
                 'L': os.path.join(input_path, 'irtf/L_fits_091201.tar'),
                 'T': os.path.join(input_path, 'irtf/T_fits_091201.tar')}

    data_folder = {'F': os.path.join(input_path, 'irtf/F_fits_091201'),
                   'G': os.path.join(input_path, 'irtf/G_fits_091201'),
                   'K': os.path.join(input_path, 'irtf/K_fits_091201'),
                   'M': os.path.join(input_path, 'irtf/M_fits_091201'),
                   'L': os.path.join(input_path, 'irtf/L_fits_091201'),
                   'T': os.path.join(input_path, 'irtf/T_fits_091201')}

    data_type = {'F': 'F stars (4.4 MB)',
                 'G': 'G stars (5.6 MB)',
                 'K': 'K stars (5.5 MB)',
                 'M': 'M stars (7.5 MB)',
                 'L': 'L dwarfs (850 kB)',
                 'T': 'T dwarfs (100 kB)'}

    url_root = 'http://irtfweb.ifa.hawaii.edu/~spex/IRTF_Spectral_Library/Data/'

    url = {'F': url_root+'F_fits_091201.tar',
           'G': url_root+'G_fits_091201.tar',
           'K': url_root+'K_fits_091201.tar',
           'M': url_root+'M_fits_091201.tar',
           'L': url_root+'L_fits_091201.tar',
           'T': url_root+'T_fits_091201.tar'}

    for item in sptypes:
        if not os.path.isfile(data_file[item]):
            print(f'Downloading IRTF Spectral Library - {data_type[item]}...', end='', flush=True)
            urllib.request.urlretrieve(url[item], data_file[item])
            print(' [DONE]')

    print('Unpacking IRTF Spectral Library...', end='', flush=True)

    for item in sptypes:
        tar = tarfile.open(data_file[item])
        tar.extractall(path=datadir)
        tar.close()

    print(' [DONE]')

    database.create_group('spectra/irtf')

    for item in sptypes:
        for root, _, files in os.walk(data_folder[item]):

            for _, filename in enumerate(files):
                if filename[-9:] != '_ext.fits':
                    fitsfile = os.path.join(root, filename)
                    spdata, header = fits.getdata(fitsfile, header=True)

                    name = header['OBJECT']
                    sptype = header['SPTYPE']

                    if name[-2:] == 'AB':
                        name = name[:-2]
                    elif name[-3:] == 'ABC':
                        name = name[:-3]

                    spt_split = sptype.split()

                    if item in ['L', 'T'] or spt_split[1][0] == 'V':
                        print_message = f'Adding IRTF Spectral Library... {name}'
                        print(f'\r{print_message:<80}', end='')

                        simbad_id, distance = queries.get_distance(name)  # [pc]

                        sptype = data_util.update_sptype(np.array([sptype]))[0]

                        dset = database.create_dataset('spectra/irtf/'+name,
                                                       data=spdata)

                        dset.attrs['name'] = str(name).encode()
                        dset.attrs['sptype'] = str(sptype).encode()
                        dset.attrs['simbad'] = str(simbad_id).encode()
                        dset.attrs['distance'] = distance

    print_message = 'Adding IRTF Spectral Library... [DONE]'
    print(f'\r{print_message:<80}')

    database.close()
